// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/filetypedialog.h"
#include "ui_filetypedialog.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/initialpos.h"
#include "OtherWssi/wssiproto.h"

FiletypeQueryDialog::FiletypeQueryDialog(QWidget *parent)
	: QMainWindow(parent), ftq(new Ui::FiletypeQueryDialog)
	{
	ftq->setupUi(this);
	this->setWindowFlags(MyWindowFlags);
	restoreSettings();
	ftq->pushButton_query->setDefault(true);
	ftq->radioButton_new_query->setChecked(true);
	connect(this, SIGNAL(signal_updateDisplay(bool)),
		mainWindowPtr, SLOT(slot_updateDisplay(bool)));
	connect(ftq->pushButton_close, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonCloseClicked()));
	connect(ftq->pushButton_query, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonQueryClicked()));
	connect(this, SIGNAL(signal_keyToTable(int,int)),
		mainWindowPtr, SLOT(slot_keyToTable(int,int)));
	connect(this, SIGNAL(signal_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)),
		mainWindowPtr, SLOT(slot_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)));
	connect(mainWindowPtr, SIGNAL(signal_close_queries()),
			  this, SLOT(close()));

// When any checkbox changes state, Query button may need to be enabled or disabled
	QList<QCheckBox*> checkboxes = this->findChildren<QCheckBox*>(QRegularExpression(QFA("checkBox*")));
	foreach (QCheckBox* checkbox, checkboxes)
		connect(checkbox, SIGNAL(stateChanged(int)), this, SLOT(slot_checkboxChanged(int)));
	this->setAttribute(Qt::WA_AlwaysShowToolTips);
	slot_checkboxChanged(0);
	}

FiletypeQueryDialog::~FiletypeQueryDialog()
	{
	delete ftq;
	}

void FiletypeQueryDialog::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("FiletypeQuerySize"))
		{
		QSize size= settings.value("FiletypeQuerySize").toSize();
		resize(size);
		}
	if(settings.contains("FiletypeQueryPos"))
		{
		QPoint pos = settings.value("FiletypeQueryPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
	else
		{
// frameGeometry is the same as geometry until the dialog is displayed
		this->show();
		move(mainWindowPtr->offCenter(this->frameGeometry(), FILETYPE_OFFSET));
		}
	}

// ------------------------------ Action buttons ------------------------------
void FiletypeQueryDialog::slot_pushButtonCloseClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("FiletypeQueryPos", pos());
	settings.setValue("FiletypeQuerySize", size());
	this->close();
	}

void FiletypeQueryDialog::slot_pushButtonQueryClicked()
	{
	if(0 == Diskno)
		return;
	for(int i=0; i<SEL_DIM; i++)
		Usesel[i] = false;
	Nselect = 0;
// If not a new query, mark which files were previously selected
	if(!ftq->radioButton_new_query->isChecked())
		updatePreviouslySelected();

	if(ftq->radioButton_add_to->isChecked())
		{
		Sel[Nselect++] = SEL_ADD_TO_PREVIOUS;
		Usesel[SEL_ADD_TO_PREVIOUS] = true;
		}

	if(ftq->radioButton_pick_from->isChecked())
		{
		Sel[Nselect++] = SEL_FROM_PREVIOUS;
		Usesel[SEL_FROM_PREVIOUS] = true;
		}

	if(ftq->radioButton_exclude_from->isChecked())
		{
		Sel[Nselect++] = SEL_REM_FROM_PREVIOUS;
		Usesel[SEL_REM_FROM_PREVIOUS] = true;
		}


	if(ftq->checkBox_fonts->isChecked())
		{
		Sel[Nselect++] = SEL_FONT;
		}
	if(ftq->checkBox_graphics->isChecked())
		{
		Sel[Nselect++] = SEL_GRAPHICS;
		}
	if(ftq->checkBox_movies->isChecked())
		{
		Sel[Nselect++] = SEL_MOVIES;
		}
	if(ftq->checkBox_self_extracting->isChecked())
		{
		Sel[Nselect++] = SEL_SELF_EXTRACTING;
		}
	if(ftq->checkBox_sound->isChecked())
		{
		Sel[Nselect++] = SEL_SOUND;
		}
	mainWindowPtr->doSelection();
	Q_EMIT signal_updateDisplay(true);
	}

// ------------------------------ Event handlers ------------------------------
// Some checkbox has changed state.  If any are enabled, query button should be enabled
void FiletypeQueryDialog::slot_checkboxChanged(int /* newstate */)
	{
	QList<QCheckBox*> checkboxes = this->findChildren<QCheckBox*>(QRegularExpression(QFA("checkBox*")));
	foreach (QCheckBox* checkbox, checkboxes)
		{
		if(checkbox->isChecked())
			{
			ftq->pushButton_query->setEnabled(true);
			ftq->pushButton_query->setToolTip(QFA(""));
			return;
			}
		}
	ftq->pushButton_query->setEnabled(false);
	ftq->pushButton_query->setToolTip(QFA("<html><head/><body><p>Check at least one file type to enable this query button.</p></body></html>"));
	}

void FiletypeQueryDialog::closeEvent(QCloseEvent *event)
	{
	event->accept();
	}

void FiletypeQueryDialog::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
		case Qt::Key_Home:
		case Qt::Key_End:
			if(!(Qt::ControlModifier & event->modifiers()))
				break;
			Q_FALLTHROUGH();
		case Qt::Key_PageDown:
		case Qt::Key_PageUp:
			Q_EMIT signal_keyToTable(event->key(), static_cast<int>(event->modifiers()));
			break;
		case Qt::Key_Escape:
			this->close();
			break;
// Intercepting keyPressEvents seems to break default button activation, so
// do it here.  Also take care setting focus after other buttons are clicked.
		case Qt::Key_Enter:
		case Qt::Key_Return:
			slot_pushButtonQueryClicked();
			break;
		default:
			return;
		}
	event->ignore();
	return;
	}

void FiletypeQueryDialog::wheelEvent(QWheelEvent *event)
	{
	Q_EMIT signal_wheelToTable(event->pos(), event->delta(), event->buttons(),
										event->modifiers(), event->orientation());
	event->ignore();
	return;
	}

