// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/indexwindow.h"
#include "OtherWssi/globalsettings.h"
#include "Dialogs/indextimefilter.h"
#include "OtherWssi/wssiproto.h"

// When filtering directories, there is the possibility that the directory is
// excluded, but an included directory is somewhere below it.  So possible
// return values are skip all, skip files or skip none.
// Counts of number of directories skipped or specifically included are incremented.
// When called, dirname does not include leading drive letter and colon, directory
// separator is / and there is are both leading and trailing slashes.  For the
// root directory, dirname is an empty QString.
DirSkipVal IndexWindow::indexTimeDirectoryFilteringSkip(QString dirname)
	{
	if(!IndexTimeFiltering)
		return SkipNone;
	if(!IndexTimeDirExclude)
		return SkipNone;
	if(0 == ((Nofile+skippedDirCount) & 256))
		qApp->processEvents();
	if(IndexTimeDirInclude)
		{
		for(int i=0; i<IndexDirIncludeList.size(); i++)
			{
			if(dirname == IndexDirIncludeList[i])
				return SkipNone;
			}
		for(int i=0; i<IndexDirIncludeList.size(); i++)
			{
			if(dirname.startsWith(IndexDirIncludeList[i], Qt::CaseInsensitive))
				return SkipNone;
			if(QString(IndexDirIncludeList[i]).startsWith(dirname, Qt::CaseInsensitive))
				break;
			}
		}
	for(int i=0; i<IndexDirExcludeList.size(); i++)
		{
		if(match(IndexDirExcludeList[i], dirname))
			return SkipAll;
		}
	return SkipNone;
	}

// Returns true if file should be skipped because of index-time editing rules.
// Counts of number of files skipped or specifically included are incremented.
bool IndexWindow::indexTimeFileFilteringSkip(QString filename)
	{
	if(0 == ((Nofile+skippedFileCount) & 256))
		qApp->processEvents();
	if(!IndexTimeFiltering)
		return false;
	if(!IndexTimeFileExclude)
		return false;
	for(int i=0; i<IndexFileExcludeList.size(); i++)
		{
		if(match(IndexFileExcludeList[i], filename))
			{
			if(IndexTimeFileInclude)
				{
				for(int j=0; j<IndexFileIncludeList.size(); j++)
					{
					if(match(IndexFileIncludeList[j], filename))
						{
						almostSkippedFileCount++;
						return false;
						}
					}
				}
			skippedFileCount++;
			return true;
			}
		}
	return false;
	}

void IndexWindow::saveIndexTimeFilter()
	{
	QSettings options("WSS-DDC", "Wssindex\\Options\\IndexTimeFiltering");
	options.setValue("IndexTimeFileExclude", IndexTimeFileExclude);
	options.setValue("IndexTimeFileInclude", IndexTimeFileInclude);
	options.setValue("IndexTimeDirExclude", IndexTimeDirExclude);
	options.setValue("IndexTimeDirInclude", IndexTimeDirInclude);

	int exlistCount = IndexFileExcludeList.size();
	if(1 == exlistCount && 0 == IndexFileExcludeList[0].size())
		exlistCount = 0;
	options.setValue("ExcludeFileListCount", exlistCount);
	for(int i=0; i<exlistCount; i++)
		options.setValue(QString("ExcludeFileList[%1]").arg(i), IndexFileExcludeList[i]);
	int inclistCount = IndexFileIncludeList.size();
	if(1 == inclistCount && 0 == IndexFileIncludeList[0].size())
		inclistCount = 0;
	options.setValue("IncludeFileListCount", inclistCount);
	for(int i=0; i<inclistCount; i++)
		options.setValue(QString("IncludeFileList[%1]").arg(i), IndexFileIncludeList[i]);
// Clean out any old include/exclude values
// Limit of 10,000 is just to stop loop if something goes awry
	for(int i=exlistCount; i<10000; i++)
		{
		if(options.contains(QString("ExcludeFileList[%1]").arg(i)))
			options.remove(QString("ExcludeFileList[%1]").arg(i));
		else
			break;
		}
	for(int i=inclistCount; i<10000; i++)
		{
		if(options.contains(QString("IncludeFileList[%1]").arg(i)))
			options.remove(QString("IncludeFileList[%1]").arg(i));
		else
			break;
		}

	exlistCount = IndexDirExcludeList.size();
	if(1 == exlistCount && 0 == IndexDirExcludeList[0].size())
		exlistCount = 0;
	options.setValue("ExcludeDirListCount", exlistCount);
	for(int i=0; i<exlistCount; i++)
		options.setValue(QString("ExcludeDirList[%1]").arg(i), IndexDirExcludeList[i]);
	inclistCount = IndexDirIncludeList.size();
	if(1 == inclistCount && 0 == IndexDirIncludeList[0].size())
		inclistCount = 0;
	options.setValue("IncludeDirListCount", inclistCount);
	for(int i=0; i<inclistCount; i++)
		options.setValue(QString("IncludeDirList[%1]").arg(i), IndexDirIncludeList[i]);
// Clean out any old include/exclude values
// Limit of 10,000 is just to stop loop if something goes awry
	for(int i=exlistCount; i<10000; i++)
		{
		if(options.contains(QString("ExcludeDirList[%1]").arg(i)))
			options.remove(QString("ExcludeDirList[%1]").arg(i));
		else
			break;
		}
	for(int i=inclistCount; i<10000; i++)
		{
		if(options.contains(QString("IncludeDirList[%1]").arg(i)))
			options.remove(QString("IncludeDirList[%1]").arg(i));
		else
			break;
		}
	}

void IndexWindow::slot_configExcludeClicked()
	{
	IndexTimeFilter *ite = new IndexTimeFilter(this);
	ite->setAttribute(Qt::WA_DeleteOnClose);
	ite->show();
	}
