// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "Dialogs/fadeanimation.h"
#include "OtherWssi/globalsettings.h"
#include "ModelView/volcombomodel.h"
#include "OtherWssi/wssiproto.h"

#include <QProgressDialog>
#include <QApplication>
#include <QMenu>

bool MainWindow::openDatabase(QString db_name)
	{
	database_name = db_name;
	if(getwss())
		{
		selectAllUnsorted();

		QSettings settings("WSS-DDC", "Wssindex");
		QString lastdb = database_name;
		QRegExp slash("/");
		int lastsep = slash.lastIndexIn(lastdb);
		if(lastsep > 0)
			{
			lastdb = lastdb.left(lastsep);
			settings.setValue("LastDatabaseDir",	lastdb);
			}
		Q_EMIT signal_diskListChanged();
		slot_updateDisplay(true);
		return true;
		}
	return false;
	}

void MainWindow::openDbFromDir(QString dbdir)
	{
// QFileDialog has a performance problem with non-native dialogs, as in
// 6 seconds to open depending on the path.  So open at root and change
// to the last directory successfully used.  This is triggered by broken A drive
// Qt 4 used its own dialog here while Qt 5 uses the native Windows dialog.
	QFileDialog *fdialog = new QFileDialog(this, tr("Open database"), "/");
	if(dbdir.size() > 0)
		fdialog->setDirectory(dbdir);
	fdialog->setViewMode(QFileDialog::Detail);
	fdialog->setFileMode(QFileDialog::ExistingFile);
	QStringList filters;
	filters << tr("Wssindex databases (*.dir)")
				<< tr("Database backups (*.bak)")
				<< tr("All files (*)");
	fdialog->setNameFilters(filters);
	fdialog->setMinimumWidth(750);

	if(fdialog->exec())
		{
		if(openDatabase(fdialog->selectedFiles().at(0)))
			{
			recentFiles->add(fdialog->selectedFiles().at(0));
			recentDirs->add(fdialog->directory().absolutePath());
			updateRecentLists();
			}
		}
	delete fdialog;
	return;
	}

void MainWindow::openDbFromFile(QString dbname)
	{
	if(openDatabase(dbname))
		{
		recentFiles->add(QDir::fromNativeSeparators(dbname));
		QFileInfo qf = QFileInfo(dbname);
		recentDirs->add(qf.absolutePath());
		}
	}

// quoteout(QString, QTextStream) - output for export a string which may contain
//	quotes or commas or other characters which would confuse an import routine.
// The comma between fields is only output by this routine if need_comma is true.

//	Special cases:
//	1. If string is a null pointer, or *string is null, do nothing
//	2. If string contains a quote, replace it with 2 quotes

void MainWindow::quoteout(QString string, QTextStream *csvout, bool *need_comma)
	{
	if(*need_comma)
		*csvout << ",";
	*need_comma = true;

	if(string.isEmpty())
		{
		*csvout << "\"\"";
		return;
		}

	*csvout << "\"";

	for(int i=0; i<string.size(); ++i)
		{
		if(string.at(i) == QLatin1Char('"'))
			*csvout << "\"";
		*csvout << string.at(i);
		}

	*csvout << "\"";

	return;
	}

// Returns true on success
bool MainWindow::saveDatabase(QString db_name)
	{
	if(0 == Nofile)
		return false;
	if(putwss(db_name))
		{
		QSettings settings("WSS-DDC", "Wssindex");
		QString lastdb = database_name;
		QRegExp slash("/");
		int lastsep = slash.lastIndexIn(lastdb);
		if(lastsep > 0)
			{
			lastdb = lastdb.left(lastsep);
			settings.setValue("LastDatabaseDir", lastdb);
			}
		return true;
		}
	return false;
	}

bool MainWindow::saveDbFromDir(QString dbdir)
	{
	if(0 == Nofile)
		{
// Avoid a clang memory leak warning, FadeAnimation deletes itself on close
#ifndef __clang_analyzer__
		new FadeAnimation(tr("Nothing to save"), tr("Sorry, an empty database cannot be saved."), Fade_time);
#endif
		return false;
		}

	QFileDialog *fdialog = new QFileDialog(this, tr("Save database"), tr("/"));
	fdialog->setMinimumWidth(750);
	fdialog->setViewMode(QFileDialog::Detail);
	fdialog->setFileMode(QFileDialog::AnyFile);
	fdialog->setWindowModality(Qt::WindowModal);
	fdialog->setAcceptMode(QFileDialog::AcceptSave);
	QStringList filters;
	filters << tr("Wssindex database (*.dir)")
			  << tr("All files (*.*)");
	fdialog->setNameFilters(filters);
	if(dbdir.size() > 0)
		fdialog->setDirectory(dbdir);
	fdialog->selectFile(database_name);
	fdialog->setLabelText(QFileDialog::Accept, tr("Save"));

	bool save_ok = false;
	if(fdialog->exec())
		{
		save_ok = saveDatabase(fdialog->selectedFiles().at(0));
		if(save_ok)
			{
			recentFiles->add(QDir::fromNativeSeparators(database_name));
			recentDirs->add(fdialog->directory().absolutePath());
			updateRecentLists();
			}
		}
	delete fdialog;
	return save_ok;
	}

void MainWindow::slot_closeClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("MainGeometryPos", pos());
	settings.setValue("MainGeometrySize", size());
	saveTopLevelSettings();

	QSettings recent("WSS-DDC", "Wssindex\\Recent");
	recent.clear();
	for(int i=0; i<NrecentFiles; i++)
		{
		recent.setValue(tr("recent_files[%1]").arg(i), recentFiles->text(i));
		recent.setValue(tr("recent_files_pinned[%1]").arg(i), recentFiles->pinned(i));
		}
	for(int i=0; i<NrecentDirs; i++)
		{
		recent.setValue(tr("recent_dirs[%1]").arg(i), recentDirs->text(i));
		recent.setValue(tr("recent_dirs_pinned[%1]").arg(i), recentDirs->pinned(i));
		}

	this->close();
	}

void MainWindow::slot_closeDbClicked()
	{
	if(Changed)
		{
		QMessageBox msgBox;
		msgBox.setText(tr("<html><head/><body><p>The database has been modified.<p>"
			"Do you want to save your changes before closing?</p></body></html>"));
		msgBox.setStandardButtons(QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
		msgBox.setDefaultButton(QMessageBox::Cancel);
		int ret = msgBox.exec();
		switch (ret)
			{
			case QMessageBox::Save:
				if(slot_saveClicked())
					break;
				Q_FALLTHROUGH();
			case QMessageBox::Cancel:
				return;
			case QMessageBox::Discard:
				break;
			default:
// should never be reached
				break;
			}
		}

	database_name = QString();
	Vcm->diskListValid = false;
	Ftm->setFileModelValidity(false);
	Q_EMIT signal_diskListChanged();
	freeall();
	slot_updateDisplay(true);
	statusNoDbLoaded();
//	ui->statusBar->showMessage(tr("No database loaded"));
//	statusMessage_db->clear();
	}

// Copy file names to clipboard
void MainWindow::slot_filenameClipboardClicked()
	{
	if(40+Noselect > INT_MAX)
		{
		QMessageBox::warning(this, tr("Can't save"), tr("Too many items are selected to be saved to the clipboard"));
		return;
		}
	QProgressDialog *progress = new QProgressDialog(this);
	progress->setLabelText(tr("Copying to clipboard..."));
	progress->setModal(true);
	progress->setCancelButton(nullptr);
	progress->setRange(0, static_cast<int>(Noselect));
	progress->setMinimumDuration(0);
	progress->show();
	qApp->processEvents();

	QClipboard *clipboard = QApplication::clipboard();
	QString *clip = new QString();
// Preallocating space for clip speeds up the process.
// We don't know the average file name length, so 40 is just a guess.
	clip->reserve(static_cast<int>(40*Noselect));
	for(unsigned long i=0; i<Noselect; i++)
		{
		unsigned long k = Sortindx[i];
		struct filerec *dr_k = Dr + k;
		if(isDotID(dr_k))
			continue;
		*clip += dr_k->Lfn + "\n";
		if(0 == (i & 4096))
			progress->setValue(static_cast<int>(i));
		}
	clipboard->setText(*clip, QClipboard::Clipboard);
	delete clip;
	progress->cancel();
	delete progress;
// Avoid a clang memory leak warning, FadeAnimation deletes itself on close
#ifndef __clang_analyzer__
	new FadeAnimation(tr("Complete"), tr("Displayed file names have been copied to the clipboard."), Fade_time);
#endif
	}

void MainWindow::slot_openClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex");
	QString dbdir = settings.value("LastDatabaseDir", "").toString();
	openDbFromDir(dbdir);
	}

void MainWindow::slot_openRecentDirClicked(int menuid)
	{
	openDbFromDir(recentDirs->text(menuid));
	}

void MainWindow::slot_openRecentFileClicked(int menuid)
	{
	openDbFromFile(recentFiles->text(menuid));
	}

void MainWindow::slot_pinRecentDirClicked(int menuid)
	{
	recentDirs->togglePin(menuid);
	updateRecentLists();
	}

void MainWindow::slot_pinRecentFileClicked(int menuid)
	{
	recentFiles->togglePin(menuid);
	updateRecentLists();
	}

bool MainWindow::slot_saveClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex");
	QString dbdir = settings.value("LastDatabaseDir", "").toString();
	return saveDbFromDir(dbdir);
	}

void MainWindow::slot_saveRecentDirClicked(int menuid)
	{
	saveDbFromDir(recentDirs->text(menuid));
	}

// xx Recently used lists need to be ordered with pinned items first.
// xx If a pin is toggled, the list needs to be resorted.  Given that
// xx the lists can have at most 10 entries, a bad implementation of
// xx bubble sort suffices.

// Changes to recentDir/File lists need to be propagated to menus
void MainWindow::updateRecentLists()
	{
	for(int i=0; i<NrecentFiles; i++)
		{
		if(recentFileOpenActs.size() < i+1)
			{
			recentFileOpenActs.append(new PinMenuItem(fileMenu, i));
			openRecentFileMenu->addAction(recentFileOpenActs[i]);
			connect(recentFileOpenActs[i], SIGNAL(signal_textClicked(int)), this, SLOT(slot_openRecentFileClicked(int)));
			connect(recentFileOpenActs[i], SIGNAL(signal_pinClicked(int)), this, SLOT(slot_pinRecentFileClicked(int)));
			}
		recentFileOpenActs[i]->setText(QDir::toNativeSeparators(recentFiles->text(i)));
		recentFileOpenActs[i]->setVisible(!recentFiles->text(i).isEmpty());
		recentFileOpenActs[i]->setPinState(recentFiles->pinned(i));
		if(recentFileOpenTBActs.size() < i+1)
			{
			recentFileOpenTBActs.append(new PinMenuItem(pickFileTB, i));
			openRecentFileTB->addAction(recentFileOpenTBActs[i]);
			connect(recentFileOpenTBActs[i], SIGNAL(signal_textClicked(int)), this, SLOT(slot_openRecentFileClicked(int)));
			connect(recentFileOpenTBActs[i], SIGNAL(signal_pinClicked(int)), this, SLOT(slot_pinRecentFileClicked(int)));
			}
		recentFileOpenTBActs[i]->setText(QDir::toNativeSeparators(recentFiles->text(i)));
		recentFileOpenTBActs[i]->setVisible(!recentFiles->text(i).isEmpty());
		recentFileOpenTBActs[i]->setPinState(recentFiles->pinned(i));
		}
// If the configured number of recently used files or directories to save
// has been reduced, there may be extra ones to hide.
	for(int i=NrecentFiles; i<recentFileOpenActs.size(); i++)
		recentFileOpenActs[i]->setVisible(false);
	for(int i=NrecentFiles; i<recentFileOpenTBActs.size(); i++)
		recentFileOpenTBActs[i]->setVisible(false);

	for(int i=0; i<NrecentDirs; i++)
		{
		if(recentDirOpenActs.size() < i+1)
			{
			recentDirOpenActs.append(new PinMenuItem(fileMenu, i));
			openRecentDirMenu->addAction(recentDirOpenActs[i]);
			connect(recentDirOpenActs[i], SIGNAL(signal_textClicked(int)), this, SLOT(slot_openRecentDirClicked(int)));
			connect(recentDirOpenActs[i], SIGNAL(signal_pinClicked(int)), this, SLOT(slot_pinRecentDirClicked(int)));
			}
		recentDirOpenActs[i]->setText(QDir::toNativeSeparators(recentDirs->text(i)));
		recentDirOpenActs[i]->setVisible(!recentDirs->text(i).isEmpty());
		recentDirOpenActs[i]->setPinState(recentDirs->pinned(i));

		if(recentDirOpenTBActs.size() < i+1)
			{
			recentDirOpenTBActs.append(new PinMenuItem(pickFileTB, i));
			openRecentDirTB->addAction(recentDirOpenTBActs[i]);
			connect(recentDirOpenTBActs[i], SIGNAL(signal_textClicked(int)), this, SLOT(slot_openRecentDirClicked(int)));
			connect(recentDirOpenTBActs[i], SIGNAL(signal_pinClicked(int)), this, SLOT(slot_pinRecentDirClicked(int)));
			}
		recentDirOpenTBActs[i]->setText(QDir::toNativeSeparators(recentDirs->text(i)));
		recentDirOpenTBActs[i]->setVisible(!recentDirs->text(i).isEmpty());
		recentDirOpenTBActs[i]->setPinState(recentDirs->pinned(i));

		if(recentDirSaveActs.size() < i+1)
			{
			recentDirSaveActs.append(new PinMenuItem(fileMenu, i));
			saveRecentDirMenu->addAction(recentDirSaveActs[i]);
			connect(recentDirSaveActs[i], SIGNAL(signal_textClicked(int)), this, SLOT(slot_saveRecentDirClicked(int)));
			connect(recentDirSaveActs[i], SIGNAL(signal_pinClicked(int)), this, SLOT(slot_pinRecentDirClicked(int)));
			}
		recentDirSaveActs[i]->setText(QDir::toNativeSeparators(recentDirs->text(i)));
		recentDirSaveActs[i]->setVisible(!recentDirs->text(i).isEmpty());
		recentDirSaveActs[i]->setPinState(recentDirs->pinned(i));

		if(recentDirSaveTBActs.size() < i+1)
			{
			recentDirSaveTBActs.append(new PinMenuItem(pickFileTB, i));
			saveRecentDirTB->addAction(recentDirSaveTBActs[i]);
			connect(recentDirSaveTBActs[i], SIGNAL(signal_textClicked(int)), this, SLOT(slot_saveRecentDirClicked(int)));
			connect(recentDirSaveTBActs[i], SIGNAL(signal_pinClicked(int)), this, SLOT(slot_pinRecentDirClicked(int)));
			}
		recentDirSaveTBActs[i]->setText(QDir::toNativeSeparators(recentDirs->text(i)));
		recentDirSaveTBActs[i]->setVisible(!recentDirs->text(i).isEmpty());
		recentDirSaveTBActs[i]->setPinState(recentDirs->pinned(i));
		}
	for(int i=NrecentDirs; i<recentDirOpenActs.size(); i++)
		recentDirOpenActs[i]->setVisible(false);
	for(int i=NrecentDirs; i<recentDirOpenTBActs.size(); i++)
		recentDirOpenTBActs[i]->setVisible(false);
	for(int i=NrecentDirs; i<recentDirSaveActs.size(); i++)
		recentDirSaveActs[i]->setVisible(false);
	for(int i=NrecentDirs; i<recentDirSaveTBActs.size(); i++)
		recentDirSaveTBActs[i]->setVisible(false);
	}
