// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/versioninfo.h"

#include <QNetworkProxyFactory>

#if defined(VER_64_BIT)
	#define BITNESS	"64"
#else
	#define BITNESS	"32"
#endif

void MainWindow::slot_about()
	{
	QMessageBox *msgBox = new QMessageBox(this);
	msgBox->setAttribute(Qt::WA_DeleteOnClose);
	msgBox->setWindowTitle(tr("About Menu"));
	msgBox->setText(tr("<h3>About Wssindex</h3>"));
	msgBox->setInformativeText(tr("<html><head/><body><p>Wssindex Disk Cataloger version %1.<br>"
#ifdef VER_64_BIT
		"(64-bit version for 64-bit Windows)<br>"
#else
		"(32-bit version for 32 or 64-bit Windows)<br>"
#endif
#ifdef DEBUG_VER
		"Debug build<br>"
#endif
#if QT_VERSION < QT_VERSION_CHECK(5, 8, 0)
	"<br>Special build compatible with XP and Vista, not recommended for any newer OS.<br>"
#endif
		"<br>%2<br><br>"
#if defined(BUILT_WITH_MSVC09)
		"Built using Microsoft Visual C++ 2008 Express"
#elif defined(BUILT_WITH_MSVC10)
		"Built using Microsoft Visual C++ 2010"
#elif defined(BUILT_WITH_MSVC11)
		"Built using Microsoft Visual C++ 2012"
#elif defined(BUILT_WITH_MSVC12)
		"Built using <a href='https://msdn.microsoft.com/en-us/library/dd831853%28v=vs.120%29.aspx'>Microsoft Visual C++ 2013</a>"
#elif defined(BUILT_WITH_MSVC14)
		"Built using <a href='https://msdn.microsoft.com/en-us/library/dd831853%28v=vs.140%29.aspx'>Microsoft Visual C++ 2015</a>"
#elif defined(BUILT_WITH_MSVC15)
		"Built using <a href='https://msdn.microsoft.com/en-us/library/dd831853%28v=vs.150%29.aspx'>Microsoft Visual C++ 2017</a>"
#elif defined(BUILT_WITH_MSVC16)
		"Built using <a href='https://msdn.microsoft.com/en-us/library/dd831853%28v=vs.150%29.aspx'>Microsoft Visual C++ 2019</a>"
#elif defined(BUILT_WITH_MinGW)
		"Built using <a href='http://www.mingw.org/'>MinGW</a>"
#else
	#error "Built-with not defined"
#endif
		" (" BITNESS "-bit) and the <a href='slotCaller:://QtAbout'>Qt toolkit</a>.<br>"
		"<br>Code has also been used from:<ul>"
		"<li><a href='http://www.info-zip.org/'>Info-Zip</a>"
		"<li><a href='https://mediaarea.net/en-us/MediaInfo'>MediaInfo</a>"
		"<li><a href='https://www.qcustomplot.com/'>QCustomPlot</a>"
		"<li><a href='https://www.mimec.org/articles/qt/wmstyle'>Windows Modern Style</a>"
		"</ul>The installer is built with <a href='https://nsis.sourceforge.io/'>The Nullsoft Scriptable Install System</a><br><br>"
		"Wssindex is licensed under the <a href='slotCaller:://Gpl3'>GNU Public License version 3</a>.<br><br>"
		"Packages used by Wssindex have <a href='slotCaller:://OtherLicenses'>additional licenses</a>.<br><br>"
		"Visit <a href='https://www.wssddc.com/wssindex/'>www.wssddc.com</a> "
		"for news, updates, source code and discussion forum.</p></body></html>")
		.arg(WSSINDEX_VERSION, WSSINDEX_COPYRIGHT));

	QPixmap pm(QLatin1String(":/Images/dvd-r-w-64.png"));
	if (!pm.isNull())
		 msgBox->setIconPixmap(pm);
	msgBox->exec();
	}

void MainWindow::slot_welcome()
	{
	welcomeWindow();
	QSettings options("WSS-DDC", "Wssindex");
	options.setValue("welcomeShown", true);
	}

// Retrieve latest version number from Wssindex home page and
// and compare to version running.
void MainWindow::slot_pushButtonVersioncheckClicked()
	{
	QNetworkAccessManager *manager = new QNetworkAccessManager(this);
// Use system proxy settings, if any
	QNetworkProxyFactory::setUseSystemConfiguration(true);
	connect(manager, SIGNAL(finished(QNetworkReply*)),
			  this, SLOT(slot_replyFinished(QNetworkReply*)));

	manager->get(QNetworkRequest(QFA(VERSION_CHECK_URL)));
	}

void MainWindow::slot_replyFinished(QNetworkReply *reply)
	{
	if(QNetworkReply::NoError == reply->error())
		{
		QIODevice *answer = reply;
// Careful, version from web may have CR/LF appended
		QString version = answer->readAll().trimmed();
		if(0 == version.compare(WSSINDEX_VERSION, Qt::CaseInsensitive))
			{
			QMessageBox::information(this, tr("Version check"),
				tr("You are running the latest version of Wssindex."));
			}
		else
			{
			QMessageBox::information(this, tr("Version check"),
				tr("<html><head/><body><p>You are running Wssindex version " WSSINDEX_VERSION
					"<br>The latest version is %1<br><br>"
					"For updates, visit <a href='https://www.wssddc.com/wssindex/"
					"#Downloads'>www.wssddc.com</a>").arg(version));
			}
		}
	else
		{
		QMessageBox::warning(this, tr("Version check failed"),
			tr("Sorry, unable to check version at this time."));
		}

	reply->deleteLater();
	}
