// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#define LIGHT_BORDER "#BBBBBB"
#define DARK_BORDER  "#999999"

#define MAX_RECENT_DIRS		10
#define MAX_RECENT_FILES	10

#include "OtherWssi/wssindex.h"
#include "OtherWssi/raise2.h"
#include "OtherWssi/pinmenu.h"
#include "OtherWssi/pinlist.h"
#include "ModelView/filetablemodel.h"
#include "ModelView/volcombomodel.h"
#include "ModelView/voltablemodel.h"
#include "Dialogs/balancecolumns.h"

#include <QMainWindow>
#include <QFile>
#include <QFileDialog>
#include <QMessageBox>
#include <QtGui>
#include <QObject>
#include <QDialog>
#include <QtNetwork/QNetworkReply>
#include <QWidgetAction>
#include <QStringListModel>
#include <QCompleter>

// Docs say that CustomizeWindowHint needs to be set to allow changing
// WindowCloseButtonHint, but that doesn't seem to be true.  Include here
// to avoid possible future breakage.  WindowTitleHint also doesn't seem
// to really be needed.
#define MyWindowFlags (Qt::CustomizeWindowHint | Qt::Dialog | Qt::WindowCloseButtonHint | Qt::WindowTitleHint)

#define BLUE_WASH "background: qlineargradient(x1: 0, y1: 0, x2: 1, y2: 1, stop: 0 #9ebef5, stop: 1.0 #a3c2fa);"

namespace Ui
	{
	class MainWindow;
	}

class settingsDialog;

class MainWindow : public QMainWindow
	{
	Q_OBJECT

public:
	MainWindow(QWidget *parent = nullptr);
	~MainWindow() override;
	Ui::MainWindow *ui;
	bool getwss();
	bool incdisk_error_check(void);
	bool incfile_and_check_for_error(bool isSubOrVol);
	bool increase_max_diskcount_to(unsigned long newmax);
	bool increase_max_filecount_to(unsigned long newmax);
	bool increase_max_subcount_to(unsigned long newmax);
	bool incsub_and_check_for_error(void);
	void purge_dups(bool silent);
	bool putwss(QString db_name);
	bool removews(QString name, unsigned long start_index);
	void set_editable_cells();
	void simpleconf(unsigned long requested_maxfiles, unsigned long requested_maxsubs,
						 unsigned long requested_maxdisks);
	void doSelection();
	void doSort(unsigned long *array, unsigned long n,
						int (*cmpf)(const unsigned long, const unsigned long), int force);
	Qt::AlignmentFlag justEnumFromInt(int flag);
	int justIntFromEnum(Qt::AlignmentFlag eflag);
	QPoint offCenter(QRect qrect, int xOffset, int yOffset);
	void updateRecentLists();
	void setAppStyle();
	void sortListPurge(void);
	void statusChanged(bool isChanged);
	void hideVolSub(bool hide);

public:
	bool update_display_on_return_to_main;
	fileTableModel *Ftm;
	QItemSelectionModel *fileSelection;
	volComboModel *Vcm;
	volTableModel *Vtm;
	settingsDialog *settings_dialog;
	QAction *helpRightSeparator;
	balanceColumns *fileBalance;
	QAction *menuHideCheckbox;
	QAction *ToolbarHideCheckbox;
	pinList *recentFiles;
	pinList *recentDirs;
	QCompleter *pastSearchesCompleter;
	QStringListModel *searchModel;
	QString appStyleTemplate;
	QString toolbarStyle;

private:
	raise2 *addDoubleIcon(QToolBar *whichBar, QIcon icon, QString tooltip1, QString tooltip2, int raise_no, const char *method, bool queue);
	QAction *addIconAction(QToolBar *whichBar, QIcon icon, QString tooltip, const char *method);
	QAction *addMenuItem(QMenu *whichMenu, QString menuText, const char *method, bool queue);
	QAction *addTopLevelMenuButton(QMenuBar *whichBar, QString menuText, const char *method);
	void addtoRecent(QString *recent, bool *pinstate, QString newitem, int maxcount);
	void checkAlpha(QColor newColor);
	void closeEvent(QCloseEvent *event) override;
	void createMenus();
	void createToolbars();
	void dragEnterEvent(QDragEnterEvent *event) override;
	void dropEvent(QDropEvent *event) override;
	void enableMenusIfDbLoaded();
	void gpl3License();
	void keyPressEvent(QKeyEvent * event) override;
	void layout_default();
	void no_files_in_db();
	unsigned long no_overflow_increment(unsigned long val, unsigned long inc, unsigned long mult);
	bool openDatabase(QString db_name);
	void openDbFromDir(QString dbdir);
	void openDbFromFile(QString dbname);
	void otherLicense();
	void quoteout(QString string, QTextStream *csvout, bool *need_comma);
	void resizeEvent(QResizeEvent *event) override;
	void restoreSettings();
	void restoreTopLevelSettings();
	bool saveDatabase(QString db_name);
	bool saveDbFromDir(QString dbdir);
	void saveTopLevelSettings();
	void setDefaultValues();
	void set_field_names();
	void sortdatabase(int force);
	void welcomeWindow();
	void setFileGridColor();
	void statusNoDbLoaded();

private:
	int headerSortDir[NFIELDS];
	int topVisibleRow;
	unsigned long file_start;

	QList<PinMenuItem *> recentFileOpenActs;
	QList<PinMenuItem *> recentFileOpenTBActs;
// Open and save use the same recent file/dir lists
	QList<PinMenuItem *> recentDirOpenActs;
	QList<PinMenuItem *> recentDirOpenTBActs;
	QList<PinMenuItem *> recentDirSaveActs;
	QList<PinMenuItem *> recentDirSaveTBActs;
	QMenu *fileMenu;
	QMenu *openRecentFileMenu;
	QMenu *openRecentDirMenu;
	QMenu *saveRecentDirMenu;
	QMenu *pickFileTB;
	QMenu *openRecentFileTB;
	QMenu *openRecentDirTB;
	QMenu *saveRecentDirTB;
	QMenu *queryMenu;
	QMenu *sortMenu;

	QAction *aboutAct;
	QAction *firstActionTB;
	QAction *lastActionTB;
	QAction *nextActionTB;
	QAction *openVollist;
	QAction *previousActionTB;
	QAction *queryAction;
	QAction *sortAction;
	QAction *tDesignAction;
	QAction *tDesignActionTB;
	QAction *useAltColorsAction;
	QAction *useAltColorsActionTB;
	QAction *useDefaultStyleAction;
	QAction *useDefaultStyleActionTB;
	QAction *useModernStyleAction;
	QAction *useModernStyleActionTB;
	QAction *vollistActionTB;
	QAction *fitWidthActionTB;
	QWidgetAction *queryActionTB;
	QWidgetAction *sortActionTB;

	QLabel *statusMessage_loaded;
	QLabel *statusMessage_db;
	QLabel *statusMessage_files;
	QLabel *statusMessage_selected;

	QString database_name;

	raise2 *fileOpenPair;
	raise2 *colorPair;
	raise2 *queryPair;
	raise2 *sortPair;
	QTimer *resizeTimer;

	struct wssicell
		{
		int row, column;
		};

	std::vector <wssicell> unused_cells;

Q_SIGNALS:
	void signal_close_queries();
	void signal_close_sorts();
	void signal_diskListChanged();
	void signal_drawPreview();
	void signal_hideSettingChanged(bool NoId);

public Q_SLOTS:
	void slot_diskListChanged();
	void slot_keyToTable(int key, int modifier);
	void slot_pushButtonInvertClicked();
	void slot_pushButtonArchivesClicked();
	void slot_pushButtonFiletypeClicked();
	void slot_pushButtonNamesClicked();
	void slot_pushButtonRangeClicked();
	void slot_pushButtonSpecialClicked();
	void slot_relayDisknameChange();
	void slot_saveSettings();
	void slot_setTableHeaders(bool update_display);
	void slot_settingsClicked();
	void slot_sortdatabase();
	void slot_sortforcedatabase(int force);
	void slot_updateDisplay(bool goToStart);
	void slot_wheelToTable(QPoint, int, Qt::MouseButtons,
										Qt::KeyboardModifiers, Qt::Orientation);
	void slot_slotCaller(QUrl url);
	void slot_recentFileMenuOpened();
	void slot_recentDirMenuOpened();
	void slot_recentFileTBOpened();
	void slot_recentDirTBOpened();
	void slot_saveRecentDirMenuOpened();
	void slot_saveRecentDirTBOpened();

private Q_SLOTS:
	bool slot_saveClicked();
	void slot_about();
	void slot_altColorsToggled();
	void slot_cellEdited(QModelIndex, QString);
	void slot_closeClicked();
	void slot_closeQueries();
	void slot_closeSorts();
	void slot_closeDbClicked();
	void slot_csvExportClicked();
	void slot_evenColorClicked();
	void slot_filenameClipboardClicked();
	void slot_oddColorClicked();
	void slot_GridColorClicked();
	void slot_openClicked();
	void slot_openIndex();
	void slot_openRecentDirClicked(int menuid);
	void slot_openRecentFileClicked(int menuid);
	void slot_openVolumes();
	void slot_pickFont();
	void slot_pickSkin();
	void slot_pinRecentDirClicked(int menuid);
	void slot_pinRecentFileClicked(int menuid);
	void slot_pushButtonVersioncheckClicked();
	void slot_pushButtonDesignClicked();
	void slot_pushButtonFirstClicked();
	void slot_pushButtonFullSortClicked();
	void slot_pushButtonKillIdClicked();
	void slot_pushButtonLastClicked();
	void slot_pushButtonNextClicked();
	void slot_pushButtonPreviousClicked();
	void slot_pushButtonQueryClicked();
	void slot_pushButtonSortClicked();
	void slot_replyFinished(QNetworkReply *reply);
	void slot_rowHeightSpinboxChanged(int);
	void slot_rowTracker(int);
	void slot_saveRecentDirClicked(int menuid);
	void slot_showallClicked();
	void slot_skinToggled();
	void slot_tableClicked(int column);
	void slot_welcome();
	void slot_resizeTimeout();
	void slot_pushButtonFitWidthClicked();
	};

#endif // MAINWINDOW_H
