// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/namesquerydialog.h"
#include "ui_namesquerydialog.h"
#include "ModelView/volcombomodel.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/initialpos.h"
#include "OtherWssi/wssiproto.h"

#include <QtDebug>

NamesQueryDialog::NamesQueryDialog(QWidget *parent)
	: QMainWindow(parent), nq(new Ui::NamesQueryDialog)
	{
	nq->setupUi(this);
	nq->comboBox_vols->setModel(mainWindowPtr->Vcm);
	nq->comboBox_vols->setCurrentIndex(-1);
	this->setWindowFlags(MyWindowFlags);
	restoreSettings();
	nq->radioButton_filename->setChecked(true);
	nq->lineEdit_filename->setFocus();
	nq->pushButton_query->setDefault(true);
	nq->radioButton_new_query->setChecked(true);
	nq->checkBox_no_dot_id->setChecked(NoId);

	mainWindowPtr->Vcm->diskListValid = false;

	connect(this, SIGNAL(signal_updateDisplay(bool)),
		mainWindowPtr, SLOT(slot_updateDisplay(bool)));
	connect(nq->pushButton_close, SIGNAL(clicked()),
		this, SLOT(slot_pushButtonCloseClicked()));

	nq->lineEdit_filename->setCompleter(mainWindowPtr->pastSearchesCompleter);

	connect(nq->pushButton_query, SIGNAL(clicked()),
			this, SLOT(slot_pushButtonQueryClicked()));
	connect(this, SIGNAL(signal_keyToTable(int,int)),
			mainWindowPtr, SLOT(slot_keyToTable(int,int)));
	connect(this, SIGNAL(signal_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)),
			mainWindowPtr, SLOT(slot_wheelToTable(QPoint,int,Qt::MouseButtons,Qt::KeyboardModifiers,Qt::Orientation)));
	connect(nq->radioButton_filename, SIGNAL(toggled(bool)),
			this, SLOT(slot_radioButtonFilenameToggled(bool)));
	connect(nq->lineEdit_filename, SIGNAL(textChanged(QString)),
			this, SLOT(slot_lineEditFilenameChanged()));

	connect(nq->radioButton_subdir, SIGNAL(toggled(bool)),
			this, SLOT(slot_radioButtonSubdirToggled(bool)));
	connect(nq->lineEdit_subdir, SIGNAL(textChanged(QString)),
			this, SLOT(slot_radioButtonSubdirChanged()));

	connect(nq->radioButton_disk, SIGNAL(toggled(bool)),
			this, SLOT(slot_radioButtonDiskToggled(bool)));
	connect(nq->comboBox_vols, SIGNAL(currentIndexChanged(int)),
			this, SLOT(slot_comboBoxVolsChanged()));

	connect(nq->radioButton_comment, SIGNAL(toggled(bool)),
			this, SLOT(slot_radioButtonCommentToggled(bool)));
	connect(nq->lineEdit_comment, SIGNAL(textChanged(QString)),
			this, SLOT(slot_lineEditCommentChanged()));

	connect(nq->radioButton_category, SIGNAL(toggled(bool)),
			this, SLOT(slot_radioButtonCategoryToggled(bool)));
	connect(nq->lineEdit_category, SIGNAL(textChanged(QString)),
			this, SLOT(slot_lineEditCategoryChanged()));

	connect(nq->radioButton_extra, SIGNAL(toggled(bool)),
			this, SLOT(slot_radioButtonExtraToggled(bool)));
	connect(nq->lineEdit_extra, SIGNAL(textChanged(QString)),
			this, SLOT(slot_lineEditExtraChanged()));

	connect(mainWindowPtr, SIGNAL(signal_close_queries()),
			this, SLOT(close()));
	connect(mainWindowPtr, SIGNAL(signal_diskListChanged()),
			this, SLOT(slot_diskListChanged()));

	connect(nq->checkBox_no_dot_id, SIGNAL(toggled(bool)),
			this, SLOT(slot_updateNoId(bool)));
	connect(mainWindowPtr, SIGNAL(signal_hideSettingChanged(bool)),
			this, SLOT(slot_updateNoId(bool)));

// Without this double setFocus, down arrow in the filename
// field triggers the completer in the simple query dialog.
	this->setFocus();
	nq->lineEdit_filename->setFocus();
	}

NamesQueryDialog::~NamesQueryDialog()
	{
	delete nq;
	}

void NamesQueryDialog::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("NamesQuerySize"))
		{
		QSize size = settings.value("NamesQuerySize").toSize();
		resize(size);
		}
	if(settings.contains("NamesQueryPos"))
		{
		QPoint pos = settings.value("NamesQueryPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
	else
		{
// frameGeometry is the same as geometry until the dialog is displayed
		this->show();
		move(mainWindowPtr->offCenter(this->frameGeometry(), NAMES_OFFSET));
		}
	}

// --------------------------------- File name --------------------------------
void NamesQueryDialog::slot_radioButtonFilenameToggled(bool checked)
	{
	if(checked)
		nq->lineEdit_filename->setFocus();
	}

void NamesQueryDialog::slot_lineEditFilenameChanged()
	{
	if(nq->lineEdit_filename->text().size() > 0)
		nq->radioButton_filename->setChecked(true);
	}

// ------------------------------- Subdirectory -------------------------------
void NamesQueryDialog::slot_radioButtonSubdirToggled(bool checked)
	{
	if(checked)
		nq->lineEdit_subdir->setFocus();
	}

void NamesQueryDialog::slot_radioButtonSubdirChanged()
	{
	if(nq->lineEdit_subdir->text().size() > 0)
		nq->radioButton_subdir->setChecked(true);
	}

// --------------------------------- Disk name --------------------------------
void NamesQueryDialog::slot_comboBoxVolsChanged()
	{
	nq->radioButton_disk->setChecked(true);
	}

void NamesQueryDialog::slot_radioButtonDiskToggled(bool checked)
	{
	if(checked)
		nq->comboBox_vols->setFocus();
	}

// ---------------------------------- Comment ---------------------------------
void NamesQueryDialog::slot_lineEditCommentChanged()
	{
	if(nq->lineEdit_comment->text().size() > 0)
		nq->radioButton_comment->setChecked(true);
	}

void NamesQueryDialog::slot_radioButtonCommentToggled(bool checked)
	{
	if(checked)
		nq->lineEdit_comment->setFocus();
	}

// --------------------------------- Category ---------------------------------
void NamesQueryDialog::slot_lineEditCategoryChanged()
	{
	if(nq->lineEdit_category->text().size() > 0)
		nq->radioButton_category->setChecked(true);
	}

void NamesQueryDialog::slot_radioButtonCategoryToggled(bool checked)
	{
	if(checked)
		nq->lineEdit_category->setFocus();
	}

// --------------------------------- Extra info -------------------------------
void NamesQueryDialog::slot_lineEditExtraChanged()
	{
	if(nq->lineEdit_extra->text().size() > 0)
		nq->radioButton_extra->setChecked(true);
	}

void NamesQueryDialog::slot_radioButtonExtraToggled(bool checked)
	{
	if(checked)
		nq->lineEdit_extra->setFocus();
	}

// ------------------------------ Action buttons ------------------------------
void NamesQueryDialog::slot_pushButtonCloseClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("NamesQueryPos", pos());
	settings.setValue("NamesQuerySize", size());
	this->close();
	}

void NamesQueryDialog::slot_pushButtonQueryClicked()
	{
	if(0 == Diskno)
		return;
	bool no_accents = nq->checkBox_ignore_diacritical->isChecked();

	NoId = nq->checkBox_no_dot_id->isChecked();
	for(int i=0; i<SEL_DIM; i++)
		Usesel[i] = false;
	Nselect = 0;
// If not a new query, mark which files were previously selected
	if(!nq->radioButton_new_query->isChecked())
		updatePreviouslySelected();

	if(nq->radioButton_add_to->isChecked())
		{
		Sel[Nselect++] = SEL_ADD_TO_PREVIOUS;
		Usesel[SEL_ADD_TO_PREVIOUS] = true;
		}

	if(nq->radioButton_pick_from->isChecked())
		{
		Sel[Nselect++] = SEL_FROM_PREVIOUS;
		Usesel[SEL_FROM_PREVIOUS] = true;
		}

	if(nq->radioButton_exclude_from->isChecked())
		{
		Sel[Nselect++] = SEL_REM_FROM_PREVIOUS;
		Usesel[SEL_REM_FROM_PREVIOUS] = true;
		}

	if(nq->radioButton_filename->isChecked())
		{
		if(no_accents)
			{
			Sel[Nselect++] = SEL_FILENAME_NODIACRIT;
			Usesel[SEL_FILENAME_NODIACRIT] = true;
			}
		else
			{
			Sel[Nselect++] = SEL_FILENAME;
			Usesel[SEL_FILENAME] = true;
			}
		QString ts = nq->lineEdit_filename->text();
		if(!ts.contains("*"))
			ts = "*" + ts + "*";
		MatchName = ts;
		}
	if(nq->radioButton_subdir->isChecked())
		{
		if(no_accents)
			{
			Sel[Nselect++] = SEL_SUBDIR_NODIACRIT;
			Usesel[SEL_SUBDIR_NODIACRIT] = true;
			}
		else
			{
			Sel[Nselect++] = SEL_SUBDIR;
			Usesel[SEL_SUBDIR] = true;
			}
		Compsub = nq->lineEdit_subdir->text().toUpper();
// Database uses "/" as path separator, but we show "\" to the user,
// so translate what the user enters.
		Compsub.replace("\\", "/");
		}
	if(nq->radioButton_disk->isChecked())
		{
		if(no_accents)
			{
			Sel[Nselect++] = SEL_VOLUME_NODIACRIT;
			Usesel[SEL_VOLUME_NODIACRIT] = true;
			}
		else
			{
			Sel[Nselect++] = SEL_VOLUME;
			Usesel[SEL_VOLUME] = true;
			}
		Compdsk = nq->comboBox_vols->currentText().toUpper();
		}
	if(nq->radioButton_comment->isChecked())
		{
		if(no_accents)
			{
			Sel[Nselect++] = SEL_COMMENT_NODIACRIT;
			Usesel[SEL_COMMENT_NODIACRIT] = true;
			}
		else
			{
			Sel[Nselect++] = SEL_COMMENT;
			Usesel[SEL_COMMENT] = true;
			}
		Compcom = nq->lineEdit_comment->text().toUpper();
		}
	if(nq->radioButton_category->isChecked())
		{
		if(no_accents)
			{
			Sel[Nselect++] = SEL_CATEGORY_NODIACRIT;
			Usesel[SEL_CATEGORY_NODIACRIT] = true;
			}
		else
			{
			Sel[Nselect++] = SEL_CATEGORY;
			Usesel[SEL_CATEGORY] = true;
			}
		Compcat = nq->lineEdit_category->text().toUpper();
		}
	if(nq->radioButton_extra->isChecked())
		{
		if(no_accents)
			{
			Sel[Nselect++] = SEL_EXTRA_INFO_NODIACRIT;
			Usesel[SEL_EXTRA_INFO_NODIACRIT] = true;
			}
		else
			{
			Sel[Nselect++] = SEL_EXTRA_INFO;
			Usesel[SEL_EXTRA_INFO] = true;
			}
		Compextra = nq->lineEdit_extra->text().toUpper();
		}
	mainWindowPtr->doSelection();
	MatchName.clear();
	updateSearchHistory(nq->lineEdit_filename->text());
	Q_EMIT signal_updateDisplay(true);
	}

void NamesQueryDialog::slot_diskListChanged()
	{
	mainWindowPtr->Vcm->diskListValid = false;
	nq->comboBox_vols->setCurrentIndex(-1);
	}

// This is called when the Exclude Subdir & Vol records checkbox changes state either
// in this dialog or in the main menu or toolbar.  Make sure all locations match.
void NamesQueryDialog::slot_updateNoId(bool checked)
	{
	NoId = checked;
	nq->checkBox_no_dot_id->setChecked(NoId);
	mainWindowPtr->menuHideCheckbox->setChecked(NoId);
	mainWindowPtr->ToolbarHideCheckbox->setChecked(NoId);
	}

// ------------------------------ Event handlers ------------------------------
void NamesQueryDialog::closeEvent(QCloseEvent *event)
	{
	event->accept();
	}

void NamesQueryDialog::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
		case Qt::Key_Home:
		case Qt::Key_End:
			if(!(Qt::ControlModifier & event->modifiers()))
				break;
			Q_FALLTHROUGH();
		case Qt::Key_PageDown:
		case Qt::Key_PageUp:
			Q_EMIT signal_keyToTable(event->key(), static_cast<int>(event->modifiers()));
			break;
		case Qt::Key_Escape:
			this->close();
			break;
// Intercepting keyPressEvents seems to break default button activation, so
// do it here.  Also take care setting focus after other buttons are clicked.
		case Qt::Key_Enter:
		case Qt::Key_Return:
			slot_pushButtonQueryClicked();
			break;
		default:
			return;
		}
	event->ignore();
	return;
	}

void NamesQueryDialog::wheelEvent(QWheelEvent *event)
	{
	Q_EMIT signal_wheelToTable(event->pos(), event->delta(), event->buttons(),
										event->modifiers(), event->orientation());
	event->ignore();
	return;
	}
