// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "Dialogs/sortbenchdialog.h"
#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include "Dialogs/mainwindow.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/initialpos.h"
#include "Workarounds/forcecompile.h"

#ifdef OLDQCP
	#include "ui_sortbenchdialog-old.h"
	#include "3rdparty/qcustomplot-old.h"
#else
	#include "ui_sortbenchdialog.h"
	#include "3rdparty/qcustomplot.h"
#endif

#include <QElapsedTimer>
#include <QDebug>

sortBenchDialog::sortBenchDialog(QWidget *parent) :
	QMainWindow(parent), sb(new Ui::sortBenchDialog)
	{
	sb->setupUi(this);
	if(Nofile < 1000)
		{
		QMessageBox::warning(this, tr("Not enough files for sort benchmarking"),
			tr("A database with at least 1000 files is needed to benchmark sort times."
				"  10,000 or more files would be better."));
		this->close();
		return;
		}
	restoreSettings();
	this->show();

	connect(this, SIGNAL(signal_sortForceNeeded(int)),
		mainWindowPtr, SLOT(slot_sortforcedatabase(int)));
	connect(sb->toolButton_close, SIGNAL(clicked()), this, SLOT(slot_pushButtonCloseClicked()));
	connect(sb->toolButton_stop, SIGNAL(clicked()), this, SLOT(slot_stopClicked()));
	connect(sb->toolButton_start, SIGNAL(clicked()), this, SLOT(slot_doTest()));
	connect(this, SIGNAL(signal_updateDisplay(bool)),
		mainWindowPtr, SLOT(slot_updateDisplay(bool)));

	statusMsg = new QLabel(tr("Ready"));
	sb->statusBar->insertWidget(0, statusMsg);

	cores = QThread::idealThreadCount();
	if(cores > 0)
		sb->label_cores->setText(tr("Detected CPU cores: %1").arg(cores));
	else
		{
		sb->label_cores->setText(tr("Unable to detect number of CPU cores, using 4 for test"));
		cores = 4;
		}

// For testing, can set arbitrary cores value here
//	cores = 32;
//	cores = qMin(cores, Max_threads);
	sb->checkBox_caseSensitive->setChecked(!Ignore_case_sort);
	sb->checkBox_natural->setChecked(NaturalSort);
	sb->tableWidget->setColumnCount(cores);
	sb->tableWidget->setRowCount(4);
	sb->tableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
	sb->tableWidget->setVerticalHeaderItem(3, new QTableWidgetItem(tr("Average")));
	sb->tableWidget->verticalHeader()->setDefaultAlignment(Qt::AlignCenter);

	coreStart = coreSave = 0;
	threadStart = threadSave = 1;
	}

sortBenchDialog::~sortBenchDialog()
	{
// We may have messed up the display by selecting all files
	Q_EMIT signal_updateDisplay(true);
	delete sb;
	}

void sortBenchDialog::slot_doTest()
	{
	mainWindowPtr->Ftm->setFileModelValidity(false);

	if(coreSave >= 2 && threadSave >= cores)
		{
		coreStart = coreSave = 0;
		threadStart = threadSave = 1;
		sb->tableWidget->clearContents();
		for(int i=0; i<3; i++)
			{
			x[i].resize(0);
			y[i].resize(0);
			}
		sb->customPlot->clearGraphs();
		}
	else
		{
		coreStart = coreSave;
		threadStart = threadSave;
		}

	Nkey = 1;
	SortKey[0] = SortKey_FN;
	Sortdir[0] = 1;
	collator->setCaseSensitivity(sb->checkBox_caseSensitive->isChecked() ? Qt::CaseSensitive : Qt::CaseInsensitive);
	collator->setNumericMode(sb->checkBox_natural->isChecked());

	if(0 == sb->customPlot->graphCount())
		{
		for(int i=0; i<3; i++)
			{
			sb->customPlot->addGraph();
			sb->customPlot->graph(i)->setLineStyle(QCPGraph::lsNone);
			sb->customPlot->graph(i)->setScatterStyle(QCPScatterStyle::ssDisc);
			}
		}
	sb->customPlot->xAxis->setRange(0, cores+0.5);
	sb->customPlot->yAxis->setRange(0, 0.1);
// Following statement would give a logarithmic y axis
// Disabled because autoscaling doesn't seem to work very well for log axes
//	sb->customPlot->yAxis->setScaleType(QCPAxis::stLogarithmic);

#ifdef OLDQCP
	sb->customPlot->xAxis->setAutoTickStep(false);
	sb->customPlot->xAxis->setTickStep(1);
	sb->customPlot->xAxis->setAutoSubTicks(false);
	sb->customPlot->xAxis->setSubTickCount(0);
#else
	sb->customPlot->xAxis->ticker()->setTickCount(cores);
#endif

	sb->customPlot->xAxis->setLabel(tr("Number of threads"));
	sb->customPlot->yAxis->setLabel(tr("Sort time (seconds)"));
	QColor tricolor[3] = {QColor(Qt::red), QColor(Qt::green), QColor(Qt::blue)};
	stopSorting = false;
	QElapsedTimer timer;
//	double avg[cores];
	double *avg = new double[static_cast<unsigned int>(cores)];
	memset(avg, 0, static_cast<size_t>(cores)*sizeof(double));
	for(int j=coreStart; j<3; j++)
		{
		int isub = 0;
		for(int i=threadStart; i<=cores; i++)
			{
			coreSave = j;
			threadSave = i;
			threadStart = 1;	// might not be 1 if resuming
			qApp->processEvents();
			selectAllUnsorted();
			if(stopSorting)
				break;
//			sb->statusBar->showMessage(tr("Sorting with %1 thread%2").arg(i).arg(i>1 ? tr("s") : tr("")));
			statusMsg->setText(tr("Sorting with %1 thread%2").arg(i).arg(i>1 ? tr("s") : tr("")));
			timer.start();
			Q_EMIT signal_sortForceNeeded(i);
			double sortsec = static_cast<double>(timer.elapsed())/1000.;
			isub++;

			x[j].resize(isub);
			y[j].resize(isub);
			x[j][isub-1] = i;
			y[j][isub-1] = sortsec*1.05;	// Force autoscaling to allow some space at the top
			sb->customPlot->graph(j)->setData(x[j], y[j]);
			sb->customPlot->graph(j)->rescaleValueAxis(true);
			sb->customPlot->graph(j)->setPen(tricolor[j]);
			y[j][isub-1] = sortsec;
			sb->customPlot->graph(j)->setData(x[j], y[j]);
			sb->customPlot->replot();
			QTableWidgetItem *newItem = new QTableWidgetItem(QString::number(sortsec, 'f', 3));
			newItem->setTextAlignment(Qt::AlignCenter);
			newItem->setForeground(tricolor[j]);
			sb->tableWidget->setItem(j, i-1, newItem);
			sb->tableWidget->resizeColumnsToContents();
			avg[i-1] += sortsec;
			}
		if(stopSorting)
			break;
		}
	statusMsg->setText(tr("Ready"));
	if(!stopSorting)
		{
		for(int i=0; i<cores; i++)
			{
			QTableWidgetItem *newItem = new QTableWidgetItem(QString::number(avg[i]/3.0, 'f', 3));
			newItem->setTextAlignment(Qt::AlignCenter);
			newItem->setForeground(QColor("black"));
			sb->tableWidget->setItem(3, i, newItem);
			}
		}
	delete [] avg;
	}

void sortBenchDialog::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("SortBenchSize"))
		{
		QSize size= settings.value("SortBenchSize").toSize();
		resize(size);
		}
	if(settings.contains("SortBenchPos"))
		{
		QPoint pos = settings.value("SortBenchPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
	else
		{
// frameGeometry is the same as geometry until the dialog is displayed
		this->show();
		move(mainWindowPtr->offCenter(this->frameGeometry(), SORTBENCH_OFFSET));
		}
	}

void sortBenchDialog::slot_stopClicked()
	{
	stopSorting = true;
	}

void sortBenchDialog::slot_pushButtonCloseClicked()
	{
// Put back the configured natural number sort setting which we may have fiddled with
	collator->setCaseSensitivity(Ignore_case_sort ? Qt::CaseInsensitive : Qt::CaseSensitive);
	collator->setNumericMode(NaturalSort);
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("SortBenchPos", pos());
	settings.setValue("SortBenchSize", size());
	this->close();
	}
