// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "Dialogs/vollistwindow.h"
//#include "OtherWssi/tablelayout.h"
#include "ui_vollistwindow.h"
//#include "ui_mainwindow.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/raise2.h"
#include "ModelView/autotooltipdelegate.h"
#include "ModelView/voltablemodel.h"
#include <QScrollBar>
#include <QSpinBox>
#include <QFontDialog>
#include <QColorDialog>
#include <QtDebug>

VollistWindow::VollistWindow(QWidget *parent)
	 : QMainWindow(parent), vw(new Ui::VollistWindow)
	{
	vw->setupUi(this);
	resizeTimer = new QTimer(this);
	this->setWindowModality(Qt::WindowModal);
	this->setWindowFlags(MyWindowFlags);
	volBalance = new balanceColumns(vw->tableView);
	restoreSettings();
	mainWindowPtr->Vtm->setTablePtr(this);
//	vw->tableView->setFont(*Font_vollist);
	filelistRedrawNeeded = false;

	topVisibleRow = 0;

	statusMessage = new QLabel;
	vw->statusBar->insertPermanentWidget(0, statusMessage);
	deleteActContext = new QAction(tr("&Delete"), this);
	deleteActContext->setStatusTip(tr("Delete volume from database"));
	connect(deleteActContext, SIGNAL(triggered()), this, SLOT(slot_deleteVolumeFromContextMenu()));

	createMenus();
	createToolbars();

	connect(vw->tableView->verticalScrollBar(), SIGNAL(valueChanged(int)),
			  this, SLOT(slot_volRowTracker(int)));

	for(int i=0; i<NVOLFIELDS; i++)
		headerVolSortDir[i] = 0;

// Narrow verticalHeader allows row resizing without showing numbers
	vw->tableView->verticalHeader()->setFixedWidth(5);
	vw->tableView->verticalHeader()->setDefaultSectionSize(Vol_row_height);

	vw->tableView->setModel(mainWindowPtr->Vtm);
	vw->tableView->setItemDelegate(new AutoToolTipDelegate(vw->tableView));
	volSelection = vw->tableView->selectionModel();
	setColHeadersAndFont();
	slot_updateVolDisplay(true);
//	volBalance->doBalance();

	vw->tableView->horizontalHeader()->setSectionsClickable(true);
	connect(vw->tableView->horizontalHeader(), SIGNAL(sectionClicked(int)),
			  this, SLOT(slot_volTableClicked(int)));
	connect(resizeTimer, SIGNAL(timeout()), this, SLOT(slot_resizeTimeout()));
	}

VollistWindow::~VollistWindow()
	{
	if(filelistRedrawNeeded)
		{
		QMetaObject::invokeMethod(mainWindowPtr, "slot_updateDisplay",
										Qt::QueuedConnection, Q_ARG(bool, true));
		}
	delete deleteActContext;

	delete closeVols;
	delete evenColorAction;
	delete matchFileColorAction;
	delete matchFileFontAction;
	delete oddColorAction;
	delete pickFontAction;
	delete purgeDupesAction;
	delete useAltColorsAction;

	delete evenColorActionTB;
	delete exitActionTB;
	delete firstActionTB;
	delete fitWidthActionTB;
	delete lastActionTB;
	delete matchFileColorActionTB;
	delete matchFileFontActionTB;
	delete nextActionTB;
	delete nullActionTB;
	delete oddColorActionTB;
	delete pickFontActionTB;
	delete previousActionTB;
	delete purgeDupesActionTB;
	delete spinActionTB;
	delete useAltColorsActionTB;
	delete colorPair;
	delete fontPair;
	delete resizeTimer;
	delete statusMessage;
	delete volBalance;
	delete vw;
	}

void VollistWindow::setColHeadersAndFont()
	{
// This is to work around an apparent Qt bug.   When the header font is changed.
// the row height adjusts to fit, but the font goes to some too-small default.
// Straightforward code would be just
//		vw->tableView->horizontalHeader()->setFont(*Font_vollist);
// Not sure if it's safe to delete the horizontal header while it's in use,
// so wait until a replacement is in place.
	QHeaderView *replacementHeader = new QHeaderView(Qt::Horizontal);
	replacementHeader->setFont(*Font_vollist);
	vw->tableView->setHorizontalHeader(replacementHeader);
	vw->tableView->horizontalHeader()->setVisible(true);

	mainWindowPtr->Vtm->setHeaderData(0, Qt::Horizontal, tr("Volume"));
	mainWindowPtr->Vtm->setHeaderData(1, Qt::Horizontal, tr("Disk size"));
	mainWindowPtr->Vtm->setHeaderData(2, Qt::Horizontal, tr("Bytes free"));
	mainWindowPtr->Vtm->setHeaderData(3, Qt::Horizontal, tr("Files"));
	mainWindowPtr->Vtm->setHeaderData(4, Qt::Horizontal, tr("Subdirectories"));
	mainWindowPtr->Vtm->setHeaderData(5, Qt::Horizontal, tr("Date indexed"));
	mainWindowPtr->Vtm->setHeaderData(6, Qt::Horizontal, tr("Category"));
	mainWindowPtr->Vtm->setHeaderData(7, Qt::Horizontal, tr("Comment"));
	mainWindowPtr->Vtm->setHeaderData(8, Qt::Horizontal, tr("Extra info"));

	volBalance->doBalance();
	}

void VollistWindow::createMenus()
	{
//	Font menu
	QMenu *pickFont = menuBar()->addMenu(tr("&Font"));
	pickFont->setStyleSheet(MENU_NO_CHECKBOX);
// Pick font
	pickFontAction = new QAction(tr("Select &font"), this);
	pickFont->addAction(pickFontAction);
	connect(pickFontAction, SIGNAL(triggered()),
			  this, SLOT(slot_volPickFont()));
// Same font as used for files
	matchFileFontAction = new QAction(tr("&Same font as used for files"), this);
	pickFont->addAction(matchFileFontAction);
	connect(matchFileFontAction, SIGNAL(triggered()),
			  this, SLOT(slot_matchFileFontClicked()));

// Row colors menu
	QMenu *pickColors = menuBar()->addMenu(tr("&Colors"));
// Alternate row colors checkbox
	useAltColorsAction = new QAction(tr("&Use alternate row colors"), this);
	useAltColorsAction->setCheckable(true);
	useAltColorsAction->setChecked(VolUseAltColors);
	pickColors->addAction(useAltColorsAction);
	connect(useAltColorsAction, SIGNAL(triggered()),
			  this, SLOT(slot_volAltColorsToggled()));
// First row color
	evenColorAction = new QAction(tr("&First row color"), this);
	pickColors->addAction(evenColorAction);
	connect(evenColorAction, SIGNAL(triggered()),
			  this, SLOT(slot_volEvenColorClicked()));
// Alternate row color
	oddColorAction = new QAction(tr("&Alternate row color"), this);
	pickColors->addAction(oddColorAction);
	connect(oddColorAction, SIGNAL(triggered()),
			  this, SLOT(slot_volOddColorClicked()));
// Grid color
	gridColorAction = new QAction(tr("&Grid color"), this);
	pickColors->addAction(gridColorAction);
	connect(gridColorAction, SIGNAL(triggered()),
			  this, SLOT(slot_volGridColorClicked()));
// Same colors as for files (includes grid color)
	matchFileColorAction = new QAction(tr("&Match colors used for files"), this);
	pickColors->addAction(matchFileColorAction);
	connect(matchFileColorAction, SIGNAL(triggered()),
			  this, SLOT(slot_matchFileColorClicked()));

// Purge duplicates
	purgeDupesAction = new QAction(tr("&Purge duplicates"), this);
	purgeDupesAction->setStatusTip(tr("Remove duplicates created by manual renaming of volumes"));
	menuBar()->addAction(purgeDupesAction);
	connect(purgeDupesAction, SIGNAL(triggered()), this, SLOT(slot_pushButtonPurgeClicked()));

// Close
	closeVols = new QAction(tr("Clos&e"), this);
	menuBar()->addAction(closeVols);
	connect(closeVols, SIGNAL(triggered()), this, SLOT(slot_pushButtonCloseClicked()));
	}

void VollistWindow::createToolbars()
	{
	vw->toolBar->setStyleSheet(mainWindowPtr->toolbarStyle);
// Font (toolbar)
	fontPair  = new raise2(this, vw->toolBar, QIcon(tr(":/Images/fontcombobox.png")),
		tr("Pick font"), tr("Select font"), SLOT(slot_volPickFont()), false);
	QMenu *pickFontTB = fontPair->menu;
	pickFontTB->setStyleSheet("QMenu::indicator {width: 1px;}");
	pickFontActionTB = new QAction(tr("Select font"), this);
	pickFontTB->addAction(pickFontActionTB);
	connect(pickFontActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_volPickFont()));
// Same font as files (toolbar pulldown)
	matchFileFontActionTB = new QAction(tr("Match font used for files"), this);
	pickFontTB->addAction(matchFileFontActionTB);
	connect(matchFileFontActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_matchFileFontClicked()));

// Toggle use of alternate colors (toolbar)
	colorPair  = new raise2(this, vw->toolBar, QIcon(tr(":/Images/rowcolors.png")),
		tr("Toggle use of alternate row colors"), tr("Set row colors"), SLOT(slot_volAltColorsToggled()), false);
	QMenu *pickColorsTB = colorPair->menu;
	useAltColorsActionTB = new QAction(tr("Use alternate row colors"), this);
	useAltColorsActionTB->setCheckable(true);
	useAltColorsActionTB->setChecked(VolUseAltColors);
	pickColorsTB->addAction(useAltColorsActionTB);
	connect(useAltColorsActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_volAltColorsToggled()));
// First row color (toolbar pulldown)
	evenColorActionTB = new QAction(tr("First row color"), this);
	pickColorsTB->addAction(evenColorActionTB);
	connect(evenColorActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_volEvenColorClicked()));
// Alternate row color (toolbar pulldown)
	oddColorActionTB = new QAction(tr("Alternate row color"), this);
	pickColorsTB->addAction(oddColorActionTB);
	connect(oddColorActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_volOddColorClicked()));
// Grid color (toolbar pulldown)
	gridColorActionTB = new QAction(tr("Grid color"), this);
	pickColorsTB->addAction(gridColorActionTB);
	connect(gridColorActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_volGridColorClicked()));
// Same color as files (toolbar pulldown)
	matchFileColorActionTB = new QAction(tr("Match colors used for files"), this);
	pickColorsTB->addAction(matchFileColorActionTB);
	connect(matchFileColorActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_matchFileColorClicked()));

// Purge duplicates (toolbar)
	purgeDupesActionTB = new QAction(QIcon(tr(":/Images/delete-duplicate.png")),
						tr("Remove duplicates created by manual renaming of volumes"), this);
	vw->toolBar->addAction(purgeDupesActionTB);
	connect(purgeDupesActionTB, SIGNAL(triggered()), this, SLOT(slot_pushButtonPurgeClicked()));

// Go to first (toolbar)
	firstActionTB = new QAction(QIcon(tr(":/Images/go-first.png")),
										 tr("Go to first disk"), this);
	vw->toolBar->addAction(firstActionTB);
	firstActionTB->setDisabled(true);
	connect(firstActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_pushButtonFirstClicked()));

// Go to previous (toolbar)
	previousActionTB = new QAction(QIcon(tr(":/Images/go-previous-24.png")),
											 tr("Go to previous group of disks"), this);
	vw->toolBar->addAction(previousActionTB);
	previousActionTB->setDisabled(true);
	connect(previousActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_pushButtonPreviousClicked()));

// Go to next (toolbar)
	nextActionTB = new QAction(QIcon(tr(":/Images/go-next-24.png")),
										tr("Go to next group of disks"), this);
	vw->toolBar->addAction(nextActionTB);
	nextActionTB->setDisabled(true);
	connect(nextActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_pushButtonNextClicked()));

// Go to last (toolbar)
	lastActionTB = new QAction(QIcon(tr(":/Images/go-last.png")),
										tr("Go to last group of disks"), this);
	vw->toolBar->addAction(lastActionTB);
	lastActionTB->setDisabled(true);
	connect(lastActionTB, SIGNAL(triggered()),
			  this, SLOT(slot_pushButtonLastClicked()));

// Fit width (toolbar)
	fitWidthActionTB = new QAction(QIcon(tr(":/Images/fitwidth.png")),
		tr("Auto-fit column widths"), this);
	vw->toolBar->addAction(fitWidthActionTB);
	connect(fitWidthActionTB, SIGNAL(triggered()),
				this, SLOT(slot_pushButtonFitVolWidthClicked()));

// Spacer (toolbar)
	nullActionTB = new QAction(QIcon(tr(":/Images/blank.png")), tr("Exit"), this);
	nullActionTB->setDisabled(true);
	vw->toolBar->addAction(nullActionTB);

// Row height spinbox (toolbar)
	QSpinBox *hspinTB = new QSpinBox(this);
	hspinTB->setRange(5,99);
	hspinTB->setValue(Vol_row_height);
	hspinTB->setPrefix(tr("Row height: "));
	hspinTB->setStatusTip(tr("Row height"));
	spinActionTB = new QWidgetAction(vw->toolBar);
	spinActionTB->setDefaultWidget(hspinTB);
	vw->toolBar->addAction(spinActionTB);
	connect(hspinTB, SIGNAL(valueChanged(int)),
			  this, SLOT(slot_rowHeightSpinboxChanged(int)));

// Spacer (toolbar)
	nullActionTB2 = new QAction(QIcon(tr(":/Images/blank.png")), tr("Exit"), this);
	nullActionTB2->setDisabled(true);
	vw->toolBar->addAction(nullActionTB2);

// Exit (toolbar)
	exitActionTB = new QAction(QIcon(tr(":/Images/exit.png")), tr("Exit"), this);
	vw->toolBar->addAction(exitActionTB);
	connect(exitActionTB, SIGNAL(triggered()), this, SLOT(slot_pushButtonCloseClicked()));
	}

void VollistWindow::restoreSettings()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	if(settings.contains("VollistGeometrySize"))
		{
		QSize size = settings.value("VollistGeometrySize").toSize();
		resize(size);
		}
	if(settings.contains("VollistGeometryPos"))
		{
		QPoint pos = settings.value("VollistGeometryPos").toPoint();
		move(pos);
		forceOnscreen(this);
		}
#define getputOption(x,y) x = options.value(#x, x).y()
	#include "OtherWssi/optionsVarlistVol.include"
#undef getputOption
	VolRowColor[0] = options.value("VolRowColor[0]", VolRowColor[0]).value<QColor>();
	VolRowColor[1] = options.value("VolRowColor[1]", VolRowColor[1]).value<QColor>();
// A bug set both colors to zero (black), same as font color, if there was no saved value.  Fix this if found.
	if((VolRowColor[0].red() + VolRowColor[0].green() + VolRowColor[0].blue() == 0 && VolRowColor[0].alpha() == 255)
	 || (VolRowColor[1].red() + VolRowColor[1].green() + VolRowColor[1].blue() == 0 && VolRowColor[1].alpha() == 255))
		{
//		slot_matchFileColorClicked();
		VolRowColor[0] = FileRowColor[0];
		VolRowColor[1] = FileRowColor[1];
		QMessageBox::information(this, tr("Bad saved colors"), tr("<html><head/><body><p>Saved background color is black, same as the font color. "
			"This is probably caused by a bug in an older version of Wssindex - on first use, no saved value exists.  The bug used "
			"zero (black).<p>Changed to use same colors as file display.  Use the red exit button to save this change."));
		}
	VolGridColor = options.value("VolGridColor", VolGridColor).value<QColor>();
	setVolGridColor();
	Font_vollist->fromString(options.value("Font_vollist", *Font_vollist).toString());
	}

// ---- Font
void VollistWindow::slot_volPickFont()
	{
// Not using QFontDialog::getFont() because Qt 5 bug disables X-to-close
// See https://bugreports.qt.io/browse/QTBUG-32240
	QFontDialog *qf = new QFontDialog(vw->tableView->font(), this);
	if(UseModernStyle)
		qf->setStyleSheet(QLatin1String("QFontDialog {" BLUE_WASH "}"));
	qf->setWindowFlags(MyWindowFlags);
	qf->setModal(true);
	qf->exec();
	if(QDialog::Accepted == qf->result())
		{
		*Font_vollist = qf->selectedFont();
		setColHeadersAndFont();
		slot_updateVolDisplay(false);
		}
	delete qf;
	}

void VollistWindow::slot_matchFileFontClicked()
	{
	*Font_vollist = *Font_filelist;
	setColHeadersAndFont();
	slot_updateVolDisplay(false);
	}

// ---- Colors
void VollistWindow::slot_volAltColorsToggled()
	{
	VolUseAltColors = !VolUseAltColors;
	useAltColorsAction->setChecked(VolUseAltColors);
	useAltColorsActionTB->setChecked(VolUseAltColors);
	slot_updateVolDisplay(false);
	}

void VollistWindow::slot_volEvenColorClicked()
	{
	QColor newColor = QColorDialog::getColor(VolRowColor[0], this,
						tr("First row color"), QColorDialog::ShowAlphaChannel);
	if(newColor.isValid())
		{
		VolRowColor[0] = newColor;
		slot_updateVolDisplay(false);
		}
	checkAlpha(newColor);
	}

void VollistWindow::slot_volOddColorClicked()
	{
	QColor newColor = QColorDialog::getColor(VolRowColor[1], this,
						tr("Alternate row color"), QColorDialog::ShowAlphaChannel);
	if(newColor.isValid())
		{
		VolRowColor[1] = newColor;
		slot_updateVolDisplay(false);
		}
	checkAlpha(newColor);
	}

void VollistWindow::slot_volGridColorClicked()
	{
	QColor newColor = QColorDialog::getColor(VolGridColor, this,
						tr("Grid color"), QColorDialog::ShowAlphaChannel);
	if(newColor.isValid())
		{
		VolGridColor = newColor;
		setVolGridColor();
		slot_updateVolDisplay(false);
		}
	checkAlpha(newColor);
	}

void VollistWindow::setVolGridColor()
	{
	vw->tableView->setStyleSheet(tr("QTableView {gridline-color: rgb(%1,%2,%3);}")
			.arg(VolGridColor.red()).arg(VolGridColor.green()).arg(VolGridColor.blue()));
	}
void VollistWindow::slot_matchFileColorClicked()
	{
	VolRowColor[0] = FileRowColor[0];
	VolRowColor[1] = FileRowColor[1];
	VolGridColor = FileGridColor;
	setVolGridColor();
	slot_updateVolDisplay(false);
	}

void VollistWindow::slot_pushButtonPurgeClicked()
	{
	mainWindowPtr->purge_dups(false);
	slot_updateVolDisplay(false);
	}

void VollistWindow::checkAlpha(QColor newColor)
	{
	if(newColor.alpha() < 10)
		QMessageBox::information(this, tr("Warning"),
		tr("The chosen background color has a small or zero alpha value which"
			" means it will be almost transparent.  If that's not what you want,"
			" change alpha to 255."));
	}

// ------------------------ Toolbar navigation routines -----------------------
void VollistWindow::slot_pushButtonFirstClicked()
	{
	volStart = 0;
	slot_updateVolDisplay(false);
	}

void VollistWindow::slot_pushButtonPreviousClicked()
	{
	if(volStart > static_cast<unsigned long>(Max_bunch))
		volStart -= static_cast<unsigned long>(Max_bunch);
	else
		volStart = 0;
	slot_updateVolDisplay(false);
	}

void VollistWindow::slot_pushButtonNextClicked()
	{
	volStart = qMin(volStart+static_cast<unsigned long>(Max_bunch), Noselect-1);
	slot_updateVolDisplay(false);
	}

void VollistWindow::slot_pushButtonLastClicked()
	{
	if(Diskno > static_cast<unsigned long>(Max_bunch))
		volStart = Diskno-static_cast<unsigned long>(Max_bunch);
	else
		volStart = 0;
	slot_updateVolDisplay(false);
	}

void VollistWindow::slot_pushButtonCloseClicked()
	{
	QSettings settings("WSS-DDC", "Wssindex\\Geometry");
	settings.setValue("VollistGeometryPos", pos());
	settings.setValue("VollistGeometrySize", size());
#define getputOption(x,y) options.setValue(#x, x)
	#define SAVING
	#include "OtherWssi/optionsVarlistVol.include"
	#undef SAVING
#undef getputOption
#define getputOption(x,y) options.setValue(#x, *x)
	getputOption(Font_vollist, dummy);
#undef getputOption
	close();
	}

void VollistWindow::slot_updateVolDisplay(bool goToStart)
	{
//	QColor rowcolor;

	if(goToStart)
		{
		volStart = 0;
		vw->tableView->horizontalHeader()->setSortIndicatorShown(false);
		}
	vw->tableView->verticalHeader()->setDefaultSectionSize(Vol_row_height);
	unsigned long vol_end = qMin(Diskno, volStart+static_cast<unsigned long>(Max_bunch));
	mainWindowPtr->Vtm->setSizeAndValidity(static_cast<int>(vol_end-volStart), NVOLFIELDS, volStart);
	if(!volBalance->colWidthsValid)
		volBalance->doBalance();

	firstActionTB->setEnabled(volStart > 0);
	previousActionTB->setEnabled(volStart > 0);
	nextActionTB->setEnabled(vol_end < Diskno);
	lastActionTB->setEnabled(vol_end < Diskno);

	vw->tableView->scrollToTop();
	vw->tableView->setFocus();

	if(!Diskno)
		return;

	if(!volBalance->colWidthsValid)
		volBalance->doBalance();

	statusMessage->setText(tr("  %L1 disks, %L2-%L3 displayed  ")
												.arg(Diskno).arg(volStart+1).arg(vol_end));
	vw->statusBar->clearMessage();
	vw->statusBar->show();
	connect(vw->tableView->model(), SIGNAL(signal_volEditCompleted(QModelIndex,QString)),
			  this, SLOT(slot_vollistCellEdited(QModelIndex,QString)));
	}

void VollistWindow::slot_rowHeightSpinboxChanged(int newvalue)
	{
	Vol_row_height = newvalue;
	vw->tableView->verticalHeader()->setDefaultSectionSize(Vol_row_height);
	vw->tableView->scrollTo(vw->tableView->model()->
			index(topVisibleRow,0), QAbstractItemView::PositionAtTop);
//	Toggling the width of the verticalHeader forces a redraw
	vw->tableView->verticalHeader()->setFixedWidth(4);
	vw->tableView->verticalHeader()->setFixedWidth(5);
// Don't want the spinbox to have focus in case user hits home/end
	vw->tableView->setFocus();
	}

void VollistWindow::slot_volRowTracker(int row)
	{
// Record what row is at the top when scroll value changes
	topVisibleRow = row;
	}

void VollistWindow::slot_volTableClicked(int column)
	{
	int localsortdir;
	if(headerVolSortDir[column] > 0)
		localsortdir = -1;
	else
		localsortdir = 1;
	headerVolSortDir[column] = localsortdir;
	Vsrtkey = column + 1;
	Vsrtdir = headerVolSortDir[column];
//	mainWindowPtr->doSort(mainWindowPtr->Vtm->vptr, Diskno, wssvcomp, 0);
	mainWindowPtr->Vtm->setVolModelValidity(false);
	mainWindowPtr->Vtm->sortDisksIfNeeded();
	slot_updateVolDisplay(true);
	Qt::SortOrder indicator;
	indicator = Vsrtdir>0 ? Qt::AscendingOrder : Qt::DescendingOrder;
	vw->tableView->horizontalHeader()->setSortIndicatorShown(true);
	vw->tableView->horizontalHeader()->setSortIndicator(column, indicator);
	}

void VollistWindow::slot_vollistCellEdited(QModelIndex index, QString newValue)
	{
	int row = index.row();
	int column = index.column();
	unsigned long volnum = mainWindowPtr->Vtm->vptr[static_cast<unsigned long>(row) + volStart];
	unsigned long first_record_on_volume = Vrec[volnum].firstOnDisk;
	struct filerec *dr_k = Dr + first_record_on_volume;
	switch(column)
		{
		case 0:		// volume name
			newValue = newValue.trimmed();
			if(newValue.isEmpty())
				{
				QMessageBox::warning(this, tr("Blank label warning"), tr("An empty volume label is not allowed"));
				slot_updateVolDisplay(false);
				break;
				}
			if((Vrec[volnum].VolName != newValue) && volumeExists(newValue))
				{
				QMessageBox msgBox;
				msgBox.setText(tr("<html><head/><body><p>New volume label already exists in database.<p>"
						"It's OK to duplicate a label if you plan on resolving the"
						" duplication by additional renaming or by purging duplicates."
						"  If you leave duplicate names in the database, nothing"
						" very bad will happen, but re-indexing will remove only the"
						" first matching name it finds.</p></body></html>"));
				msgBox.addButton(tr("Do it"), QMessageBox::ActionRole);
				QPushButton *cancelButton = msgBox.addButton(tr("Cancel"), QMessageBox::ActionRole);
				msgBox.exec();
				if(static_cast<QPushButton *>(msgBox.clickedButton()) == cancelButton)
					{
					slot_updateVolDisplay(false);
					break;
					}
				}
			Vrec[volnum].VolName = newValue;
			for(unsigned int k=first_record_on_volume; k<Nofile; k++)
				{
				dr_k = Dr + k;
				if(volnum != dr_k->Volno)
					break;
				setVolEdited(dr_k);
				}
			mainWindowPtr->statusChanged(true);
			slot_updateVolDisplay(false);
			selectAllUnsorted();
// If the main window has a selection dialog open with a pulldown for
// disk name, tell the main window to tell the dialog to refresh the list
			QMetaObject::invokeMethod(mainWindowPtr, "slot_relayDisknameChange",
											Qt::QueuedConnection);
			break;

		case 6:		// category
			if(!isDotID(dr_k))
				{
				noDotidWarning();
				break;
				}
			rtrimInPlace(newValue);
			dr_k->Category = newValue;
			setCategoryEdited(dr_k);
			if(newValue.isEmpty())
				setHasNoCategory(dr_k);
			else
				setHasCategory(dr_k);
			mainWindowPtr->statusChanged(true);
			break;

		case 7:		// comment
			if(!isDotID(dr_k))
				{
				noDotidWarning();
				break;
				}
			rtrimInPlace(newValue);
			dr_k->Comment = newValue;
			setCommentEdited(dr_k);
			if(newValue.isEmpty())
				setHasNoComment(dr_k);
			else
				setHasComment(dr_k);
			mainWindowPtr->statusChanged(true);
			break;

		case 8:		// extra info
			if(!isDotID(dr_k))
				{
				noDotidWarning();
				break;
				}
			rtrimInPlace(newValue);
			dr_k->ExtraInfo = newValue;
			setExtraEdited(dr_k);
			if(newValue.isEmpty())
				setHasNoExtra(dr_k);
			else
				setHasExtra(dr_k);
			mainWindowPtr->statusChanged(true);
			break;

		default:
			QMessageBox::warning(this, tr("Error"),
				tr("Only volume name, comment, category and extra info are editable"));
			break;
		}

	filelistRedrawNeeded = true;
	}

void VollistWindow::noDotidWarning()
	{
	QMessageBox::warning(this, tr("Warning"),
		tr("Disk does not have a <Volume> entry.  The only editable field is the volume name."));
	}

// find_volume - look for volume name in database.  If found, return pointer
//	to disk record for volume, otherwise return NULL.

bool VollistWindow::volumeExists(QString volname)
	{
	for(unsigned long j=0; j<Diskno; j++)
		{
		struct diskrec *vrec_j = Vrec + j;
		if(0 == qstringCompare(vrec_j->VolName, volname))
			return(true);
		}

	return(false);
	}

void VollistWindow::contextMenuEvent(QContextMenuEvent *event)
	{
	QPoint widgetpos = vw->tableView->viewport()->mapFromGlobal(event->globalPos());
	QModelIndex index = vw->tableView->indexAt(widgetpos);
	volContextClick = static_cast<unsigned long>(index.row());
	QMenu menu(this);
	menu.addAction(deleteActContext);
	menu.exec(event->globalPos());
	}

// Delete volume number volContextClick
void VollistWindow::slot_deleteVolumeFromContextMenu()
	{
	unsigned long vol_to_delete = mainWindowPtr->Vtm->vptr[volContextClick + volStart];
	disconnect(vw->tableView->model(), SIGNAL(signal_volEditCompleted(QModelIndex,QString)),
			  this, SLOT(slot_vollistCellEdited(QModelIndex,QString)));
	mainWindowPtr->Vtm->setRed(volContextClick + volStart);

	int ret = QMessageBox::question(this, tr("Delete disk"), tr("Delete volume %1 (highlighted)?")
				.arg(Vrec[vol_to_delete].VolName), QMessageBox::Yes | QMessageBox::No);
	mainWindowPtr->Vtm->noRed();
	if(QMessageBox::Yes == ret)
		{
		if(mainWindowPtr->removews(Vrec[vol_to_delete].VolName, vol_to_delete))
			{
			QMessageBox::critical(this, tr("Error"), tr("Unable to remove volume, something is seriously awry"));
			}
		slot_updateVolDisplay(false);
// If the main window has a selection dialog open with a pulldown for
// disk name, tell the main window to tell the dialog to refresh the list
		QMetaObject::invokeMethod(mainWindowPtr, "slot_relayDisknameChange",
										Qt::QueuedConnection);
		filelistRedrawNeeded = true;
		}
	connect(vw->tableView->model(), SIGNAL(signal_volEditCompleted(QModelIndex,QString)),
			  this, SLOT(slot_vollistCellEdited(QModelIndex,QString)));
	}

void VollistWindow::slot_pushButtonFitVolWidthClicked()
	{
	volBalance->doBalance();
	}

void VollistWindow::resizeEvent(QResizeEvent * /* event */)
	{
	resizeTimer->stop();
	resizeTimer->start(RESIZING_DELAY_MS);
	}

void VollistWindow::slot_resizeTimeout()
	{
	resizeTimer->stop();
	volBalance->colWidthsValid = false;
	volBalance->doBalance();
	}

void VollistWindow::keyPressEvent(QKeyEvent *event)
	{
	switch(event->key())
		{
// Control-D means copy previous non-empty cell, but only if selected cell
// is one of the fields where this makes sense.  If displaying as multiple
// bunches, do not look outside the current bunch.
		case Qt::Key_D:
			if(0 == Nofile || !(Qt::ControlModifier & event->modifiers()))
				return;
			{
			int row = vw->tableView->currentIndex().row();
			if(row <= 0)
				return;
			int column = vw->tableView->currentIndex().column();
			unsigned long *sortedVolPtr = mainWindowPtr->Vtm->vptr;
			unsigned long volnum = mainWindowPtr->Vtm->vptr[static_cast<unsigned long>(row) + volStart];
			unsigned long firstRecordOnVolume = Vrec[volnum].firstOnDisk;
			struct filerec *dr_k = Dr + firstRecordOnVolume;
			switch(column)
				{
				case VOL_DISP_CATEGORY:
					for(unsigned long i=1; i<=volnum-volStart; i++)
						{
						unsigned long j = Vrec[sortedVolPtr[volnum-i]].firstOnDisk;
						if(!(HAS_CATEGORY & Dr[j].Diskbits))
							continue;
						if(Dr[j].Category.isEmpty())
							continue;
						dr_k->Category = Dr[j].Category;
						setCategoryEdited(dr_k);
						setHasCategory(dr_k);
						break;
						}
					break;
				case VOL_DISP_COMMENT:
					for(unsigned long i=1; i<=volnum; i++)
						{
						unsigned long j = Vrec[sortedVolPtr[volnum-i]].firstOnDisk;
						if(!(HAS_COMMENT & Dr[j].Diskbits))
							continue;
						if(Dr[j].Comment.isEmpty())
							continue;
						dr_k->Comment = Dr[j].Comment;
						setCommentEdited(dr_k);
						setHasComment(dr_k);
						break;
						}
					break;
				case VOL_DISP_EXTRA_INFO:
					for(unsigned long i=1; i<=volnum; i++)
						{
						unsigned long j = Vrec[sortedVolPtr[volnum-i]].firstOnDisk;
						if(!(HAS_EXTRA_INFO & Dr[j].Diskbits))
							continue;
						if(Dr[j].ExtraInfo.isEmpty())
							continue;
						dr_k->ExtraInfo = Dr[j].ExtraInfo;
						setExtraEdited(dr_k);
						setHasExtra(dr_k);
						break;
						}
					break;
				default:
					return;
				}
			mainWindowPtr->statusChanged(true);
			mainWindowPtr->Ftm->refreshData();
			mainWindowPtr->Vtm->refreshData();
			event->accept();
			}
		}
	}
