// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Non_member_functions/customio.h"
#include "Dialogs/mainwindow.h"

#include <fcntl.h>
#include <io.h>

// arjadd - read an .ARJ file and add information to wssindex database.
//	If mode is 2, store only archive comment.

#define ARJ_TAG				0xEA60U		// 6000U
#define HEADER_ID_HI			0xEA
#define HEADER_ID_LO			0x60
#define CRCPOLY				0xEDB88320L
#define CRC_MASK				0xFFFFFFFFL
#define FIRST_HDR_SIZE		30
#define COMMENT_MAX			2048
#define HEADERSIZE_MAX		(FIRST_HDR_SIZE + 10 + FNAME_MAX + COMMENT_MAX)
#define MAXSFX					25000
#define FNAME_MAX				512

#define TEMPBUF_SIZE 4096

static unsigned long arj_crctable[256];

#pragma pack(1)

int IndexWindow::arjadd(QString path, QString filename, int mode)
	{
	char tempbuf[TEMPBUF_SIZE];
	struct ar1
		{
		unsigned short tag;				// 0xea60
		unsigned short basic_header_size;
		} arj_begin;

	struct ar2
		{
		char archiver_version;
		char minimum_version;
		char host_os;
		char flags;						// 1 bit = garbled flag
											//	2 bit = reserved
											//	4 bit = volume flag
											//	8-bit = continued file flag
											//	0x10  = path names translated to forward
											//				slashes, usually set
		char method;
		char file_type;				// = 2 for main header
		char reserved;
		unsigned short time;		// time stamp, DOS format
		unsigned short date;		// date stamp, DOS format
		long compressed_size;		// compressed file size (bytes)
		long uncompressed_size;	// uncompressed size
		long crc;						// cyclic redundancy check, not used here
		short entryname_position;
		short access_mode;
		short host_data;
		long extra_data;				// may not be present
		} arj_header;

	long	head_position,			// file pointer at start of header
			lastpos;
	int	comment_size,			// length of comment, if any
			extended_header_size,// length of extended header, if any
			first_header_size,
			headersize,
			temp;						// temp. used while skipping over archive name

	FILE *fp;						// file pointer for reading archive

//	qDebug() << filename;

// The file extension being processed is ARJ or A00-A99,
//	or, the extension is EXE for a potential self-extracting file.
	int return_code = EXTRACTED_OK;
	unsigned long basefile = Nofile - 1;
	struct filerec *dr_nofile = Dr + basefile;
	bool possibleSelfExtracting = false;
	bool foundContent = false;
	if(0 == QFA("e").compare(dr_nofile->Lfn.midRef(dr_nofile->ExtOffset,1), Qt::CaseInsensitive))
		{
		possibleSelfExtracting = true;
		if(skip_extract(filename))
			return(SKIPPED);
		if(nullptr == (fp = arc_open(path, filename)))
			return(CANT_OPEN);
		lastpos = _filelength(_fileno(fp)) - 2;
		if (lastpos > MAXSFX)
			lastpos = MAXSFX;
		for (long arcpos = 0; arcpos < lastpos; arcpos++)
			{
			fseek(fp, arcpos, SEEK_SET);
			unsigned char c = static_cast<unsigned char>(fgetc(fp));
			while (arcpos < lastpos)
				{
				if (c != HEADER_ID_LO)  // low order first
					c = static_cast<unsigned char>(fgetc(fp));
				else if ((c = static_cast<unsigned char>(fgetc(fp))) == HEADER_ID_HI)
					break;
				arcpos++;
				}
			if (arcpos >= lastpos)
				{
				return_code = NOT_RECOGNIZED;
				goto norm_ret;
				}
			if ((headersize = static_cast<int>(fgetWord(fp))) <= HEADERSIZE_MAX)
				{
				unsigned long crc = CRC_MASK;
				for(int i=0; i<headersize; i++)
					{
					c = static_cast<unsigned char>(fgetc(fp));
					crc = arj_crctable[(static_cast<unsigned char>(crc)^static_cast<unsigned char>(c)) & 0xff]
							^ (crc >> CHAR_BIT);
					}
				if ((crc ^ CRC_MASK) == fgetLongword(fp))
					{
					fseek(fp, arcpos, SEEK_SET);
					break;
					}
				}
			}
		}
	else
		{
		do {
			if(dr_nofile->Lfn.endsWith(QFA("arj"),Qt::CaseInsensitive))
				break;
			if(dr_nofile->Lfn.right(2)[0].isDigit() && dr_nofile->Lfn.right(1)[0].isDigit())
				break;
			return(WILDCARD_NOT_ME);
			} while(false);
		if(skip_extract(filename))
			return(SKIPPED);

		if(nullptr == (fp = arc_open(path, filename)))
			return(CANT_OPEN);
		}

	if(1 != fread(&arj_begin, sizeof(struct ar1), 1, fp)
		|| arj_begin.tag != ARJ_TAG)
		{
		goto corrupt;			// this isn't an error, just a warning
		}

	head_position = ftell(fp);
	first_header_size = fgetc(fp)-1;
	if(first_header_size <= 0)
		goto corrupt;
	if(first_header_size > static_cast<int>(sizeof(arj_header)))
		goto corrupt;

	if(1 != fread(&arj_header, static_cast<size_t>(first_header_size), 1, fp))
		goto corrupt;

// skip over archive file name to get to comment
	while(0 != (temp=fgetc(fp)) && EOF != temp) ;

// comment length includes null terminator, so should be at least 1
	comment_size = arj_begin.basic_header_size - static_cast<int>(ftell(fp)-head_position);

	if((extractArcComment & mode) && (comment_size > 1))
		{
		if(comment_size > TEMPBUF_SIZE-1)
			comment_size = TEMPBUF_SIZE-1;
		if(1 != fread(tempbuf, static_cast<size_t>(comment_size), 1, fp))
			{
			goto corrupt;
			}
		if(com_arch(tempbuf, dr_nofile, comment_size))
			goto no_core;
		}
	if(!(extractMembers & mode))
		goto norm_ret;

// In a secure arj file, the compressed file size in the first header contains
//	the size of the entire file; in ordinary arj files it contains zero.  Set
//	to zero here else seek below skips over entire file.
	arj_header.compressed_size = 0;

	for(;;)
		{
// seek past previous file comment, header CRC, any extended header and
//	compressed file data
		fseek(fp, head_position+arj_begin.basic_header_size+4, SEEK_SET);
		extended_header_size = static_cast<int>(fgetWord(fp));
		if(WSSERROR_OR_EOF(fp))
			goto corrupt;
		if(extended_header_size > 0)
			fseek(fp, 4+extended_header_size, SEEK_CUR);
		fseek(fp, arj_header.compressed_size, SEEK_CUR);

		if(1 != fread(&arj_begin, sizeof(struct ar1), 1, fp)
			|| arj_begin.tag != ARJ_TAG)
			{
			goto corrupt;			// this isn't an error, just a warning
			}
		if(0 == arj_begin.basic_header_size)
			goto norm_ret;

		head_position=ftell(fp);
		first_header_size = fgetc(fp)-1;
		if(first_header_size <= 0)
			goto corrupt;
		if(first_header_size > static_cast<int>(sizeof(arj_header)))
			goto corrupt;

		if(1 != fread(&arj_header, static_cast<size_t>(first_header_size), 1, fp))
			goto corrupt;

		if( !(4 & arj_header.flags))
			{
			dr_nofile=Dr + Nofile;
			fseek(fp, arj_header.entryname_position, SEEK_CUR);
			getNullTerminatedString(tempbuf, TEMPBUF_SIZE, fp);

			switch(Case_option_indexing)
				{
				case CASE_PRESERVE:
					break;
				case CASE_FORCE_UP:
					_strupr(tempbuf);
					break;
				case CASE_FORCE_LOW:
					_strlwr(tempbuf);
					break;
				}

			foundContent = true;
			dr_nofile->Ddate = datextract(arj_header.date);
			dr_nofile->Dtime = timextract(arj_header.time);
			dr_nofile->Dsize = static_cast<quint64>(arj_header.uncompressed_size);
			dr_nofile->Volno = static_cast<unsigned long>(Diskno);
			dr_nofile->Subdirno = static_cast<unsigned long>(basefile);
			dr_nofile->Category = QString();
			dr_nofile->Comment = QString();
			dr_nofile->Bitflags = 0;
			dr_nofile->Diskbits = 0;
			setInArc(dr_nofile);
			unbad(tempbuf, 0);
			dr_nofile->Lfn = QFA(tempbuf);
			dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
			setHasLFN(dr_nofile);

			if(extractMemberComments & mode)
				{
				getNullTerminatedString(tempbuf, TEMPBUF_SIZE, fp);
				if(EOS != *tempbuf && com_arch(tempbuf, dr_nofile, temp))
					goto no_core;
				}

			if(mainWindowPtr->incfile_and_check_for_error(false))
				goto no_core;
			}
		}

// never fall out of above loop
corrupt:
	if(foundContent || !possibleSelfExtracting)
		{
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not an ARJ archive.")
			.arg(path, filename));
		}
// fall through to norm_ret

norm_ret:
		fclose(fp);
//		update_db_totals();
		qApp->processEvents();
		return(return_code);

no_core:
		return_code = DATABASE_FULL;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in arjadd"));
		Stop_indexing = true;
		goto norm_ret;
	}

void IndexWindow::make_arj_crctable(void)
	{
	int i, j;

	for(i=0; i<=255; i++)
		{
		unsigned long r = static_cast<unsigned int>(i);
		for (j = CHAR_BIT; j > 0; j--)
			{
			if (r & 1)
				r = (r >> 1) ^ CRCPOLY;
			else
				r >>= 1;
			}
		arj_crctable[i] = r;
		}
	}
