// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Non_member_functions/customio.h"
#include "Dialogs/mainwindow.h"

// cabadd - read a .CAB file and add information to wssindex database.

//(1) CAB_HEADER structure
//(2) Reserved area, if CAB_HEADER.flags & CAB_FLAG_RESERVE
//(3) Previous cabinet name, if CAB_HEADER.flags & CAB_FLAG_HASPREV
//(4) Previous disk name, if CAB_HEADER.flags & CAB_FLAG_HASPREV
//(5) Next cabinet name, if CAB_HEADER.flags & CAB_FLAG_HASNEXT
//(6) Next disk name, if CAB_HEADER.flags & CAB_FLAG_HASNEXT
//(7) CAB_FOLDER structures (n = CAB_HEADER.cFolders)
//(8) CAB_ENTRY structures / file names (n = CAB_HEADER.cFiles)
//(9) File data (offset = CAB_FOLDER.coffCabStart)

#pragma pack(1)

#define SWAPSHORT(x)	((short) (((x) << 8) | ((x) >> 8)))
#define SWAPLONG(x)	((SWAPSHORT ((short) (x)) << 16) | \
							 (SWAPSHORT ((short) ((x) >> 16))))


#define CAB_SIGNATURE		  0x4643534dUL	// MSCF
#define ALPHA_SIGNATURE		  0x28635349UL	// ISc(
#define CAB_VERSION			  0x0103

#define CAB_FLAG_HASPREV	  0x0001
#define CAB_FLAG_HASNEXT	  0x0002
#define CAB_FLAG_RESERVE	  0x0004

#define CAB_ATTRIB_READONLY  0x0001
#define CAB_ATTRIB_HIDDEN	  0x0002
#define CAB_ATTRIB_SYSTEM	  0x0004
#define CAB_ATTRIB_VOLUME	  0x0008
#define CAB_ATTRIB_DIRECTORY 0x0010
#define CAB_ATTRIB_ARCHIVE	  0x0020

#define CAB_FILE_FIRST		  0x0000
#define CAB_FILE_NEXT		  0x0001
#define CAB_FILE_SPLIT		  0xFFFE
#define CAB_FILE_CONTINUED	  0xFFFD

#define CAB_NOTIFY_OK		  1
#define CAB_NOTIFY_ERROR	  0
#define CAB_NOTIFY_SKIP		  0
#define CAB_NOTIFY_ABORT	  (-1)

#define TEMPBUF_SIZE 4096

int IndexWindow::cabadd(QString path, QString filename, int /* dummy */)
	{
struct CAB_HEADER
	{
	long sig;					// file signature 'MSCF' (CAB_SIGNATURE)
	long csumHeader;			// header checksum (0 if not used)
	long cbCabinet;			// cabinet file size
	long csumFolders;			// folders checksum (0 if not used)
	long coffFiles;			// offset of first CAB_ENTRY
	long csumFiles;			// files checksum (0 if not used)
	short	 version;			// cabinet version (CAB_VERSION)
	short	 cFolders;			// number of folders
	short	 cFiles;				// number of files
	short	 flags;				// cabinet flags (CAB_FLAG_*)
	short	 setID;				// cabinet set id
	short	 iCabinet;			// zero-based cabinet number
	}
cab_header;
#define CAB_HEADER_SIZE sizeof (struct CAB_HEADER)

//struct CAB_FOLDER
//	{
//	long coffCabStart;		// offset of folder data
//	short	 cCFData;			// ???
//	short	 typeCompress;		// compression type (tcomp* in FDI.h)
//	}
//cab_folder;
//#define CAB_FOLDER_SIZE sizeof (struct CAB_FOLDER)

struct CAB_ENTRY
	{
	long cbFile;				// uncompressed file size
	long uoffFolderStart;	// file offset after decompression
	unsigned short	 iFolder;			// file control id (CAB_FILE_*)
	unsigned short	 date;				// file date stamp, as used by DOS
	unsigned short	 time;				// file time stamp, as used by DOS
	unsigned short	 attribs;			// file attributes (CAB_ATTRIB_*)
	}
cab_entry;
#define CAB_ENTRY_SIZE sizeof (struct CAB_ENTRY)

	char tempbuf[TEMPBUF_SIZE];
	FILE *fp;
	struct filerec *dr_nofile;
	char *full_name;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	if(1 != fread(&cab_header, CAB_HEADER_SIZE, 1, fp))
		goto corrupt;

// CAB files for Alphas are completely different.  Ignore them.
	if(ALPHA_SIGNATURE == cab_header.sig)
		goto norm_ret;

	if(CAB_SIGNATURE != cab_header.sig)
		goto corrupt;

	if(0 != fseek(fp, cab_header.coffFiles, SEEK_SET))
		goto corrupt;

	for(int i=0; i<cab_header.cFiles; i++)	// ??? what about cFolders
		{
		if(1 != fread(&cab_entry, CAB_ENTRY_SIZE, 1, fp))
			goto corrupt;
		getNullTerminatedString(tempbuf, TEMPBUF_SIZE, fp);
		dr_nofile = Dr + Nofile;
// Skip over any path information in file name
		full_name = tempbuf;
		if(!Store_arc_path)
			full_name = no_path(tempbuf, static_cast<int>(strlen(tempbuf)));

		switch(Case_option_indexing)
			{
			case CASE_PRESERVE:
				break;
			case CASE_FORCE_UP:
				_strupr(full_name);
				break;
			case CASE_FORCE_LOW:
				_strlwr(full_name);
				break;
			}

		dr_nofile->Ddate = datextract(cab_entry.date);
		dr_nofile->Dtime = timextract(cab_entry.time);
		dr_nofile->Dsize = static_cast<quint64>(cab_entry.cbFile);
		dr_nofile->Volno = Diskno;
		dr_nofile->Subdirno = basefile;
		dr_nofile->Comment = QString();
		dr_nofile->Category = QString();
		dr_nofile->Bitflags = 0;
		dr_nofile->Diskbits = 0;
		setInArc(dr_nofile);
		unbad(full_name,  0);
		dr_nofile->Lfn = full_name;
		dr_nofile->ExtOffset = static_cast<unsigned short>(lastdot(dr_nofile->Lfn));
		setHasLFN(dr_nofile);

		if(mainWindowPtr->incfile_and_check_for_error(false))
			goto no_core;

		if(CAB_FILE_SPLIT == cab_entry.iFolder)
			goto norm_ret;
		}
	goto norm_ret;

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a CAB archive.")
				.arg(path, filename));
// fall through to norm_ret

norm_ret:
		fclose(fp);
//		update_db_totals();
		qApp->processEvents();
		return(return_code);

no_core:
		return_code = DATABASE_FULL;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in cabadd"));
		Stop_indexing = true;
		goto norm_ret;
	}
