// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "ExtractSubs/extractinfo.h"
#include "stdio.h"
#include "Non_member_functions/customio.h"
#include <sstream>
#include <iomanip>

#include <QMessageBox>

// aviadd - read an AVI movie file and add information to wssindex
//	database.

// Comment, category and extra info are limited to 99 characters by the settings
// spinbox.  If these limits are increased substantially, TEMPBUF_SIZE may also
// need to be increased.
#define TEMPBUF_SIZE 256

// Read a "chunk" header from fp.
// If chunkname1 is RIFF or LIST, chunkname2 is identifier for subchunk,
// e.g., AVI or WAVE.
// If chunk is not of interest, skip forward by length to the next chunk.
// Return false on error.
bool IndexWindow::read_avi_chunk(FILE *fp, unsigned long *chunkname1, unsigned long *chunkname2, unsigned long *length)
	{
	*chunkname1 = fgetLongword(fp);
	*length = fgetLongword(fp);
	if(0x46464952UL == *chunkname1 || 0x5453494cUL == *chunkname1)	// "RIFF or "LIST"
		*chunkname2 = fgetLongword(fp);
	if(ferror(fp))
		return false;
	return true;
	}

int IndexWindow::aviadd(QString path, QString filename, int /* dummy */)
	{
	char tempbuf[TEMPBUF_SIZE];
	FILE *fp;
	struct filerec *dr_nofile;
	long run_hours, run_minutes;
	double dwScale, dwRate, dwLength;
	unsigned long chunkname1, chunkname2, length;
// height, width and run_seconds are initialized to quiet compiler warnings.
	long height = 0;
	long width = 0;
	double run_seconds = 0;
	std::stringstream ss;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);
//	qDebug() << "Processing " << filename;

	bool havetime = false;
	bool havedimensions = false;
	if(!read_avi_chunk(fp, &chunkname1, &chunkname2, &length))
		goto corrupt;
	if(0x46464952UL != chunkname1)		// "RIFF"
		goto corrupt;
	if(0x20495641UL != chunkname2)		// "AVI "
		goto corrupt;
	for(;;)
		{
		if(!read_avi_chunk(fp, &chunkname1, &chunkname2, &length))
			goto corrupt;
		switch(chunkname1)
			{
			case 0x5453494cUL:	// "LIST"
				continue;

			case 0x68727473:		// "strh"
				fseek(fp, 20, SEEK_CUR);
				dwScale = static_cast<double>(fgetLongword(fp));
				dwRate = static_cast<double>(fgetLongword(fp));
				if(0 <= dwRate)
					dwRate = 1.0;
				fseek(fp, 4, SEEK_CUR);
				dwLength = static_cast<double>(fgetLongword(fp));
				run_seconds = dwLength*dwScale/dwRate;
				havetime = true;
				break;

			case 0x68697661UL:	// "avih"
				fseek(fp, 32, SEEK_CUR);
				width = static_cast<long>(fgetLongword(fp));
				height = static_cast<long>(fgetLongword(fp));
				fseek(fp, 16, SEEK_CUR);
				havedimensions = true;
				break;

			case 0x6c726468UL:	// "hdrl"
				fseek(fp, static_cast<long>(length), SEEK_CUR);
				break;

			default:
				goto corrupt;
			}
		if(havetime && havedimensions)
			break;
		}

	run_hours = static_cast<int>(run_seconds/3600);
	run_seconds -= 3600.*run_hours;
	run_minutes = static_cast<long>(run_seconds/60);
	run_seconds -= 60.*run_minutes;
	dr_nofile = Dr + basefile;
	ss << width << "x" << height << ", " << run_hours << ":" << std::setfill('0') << std::setw(2) << run_minutes
		<< ":" << std::setfill('0') << std::setw(2)<< std::fixed << std::setprecision(0) << run_seconds;
	strncpy_s(tempbuf, TEMPBUF_SIZE, ss.str().c_str(), TEMPBUF_SIZE-1);
	if(extra_arch(tempbuf, dr_nofile, Max_extra_length))
		goto no_core;
	goto norm_ret;

corrupt:
	return_code = CORRUPT;
	Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a recognized AVI movie file.")
			.arg(path, filename));

norm_ret:
	fclose(fp);
	return(return_code);

no_core:
	return_code = DATABASE_FULL;
	Stop_indexing = true;
	QMessageBox::critical(this, tr("Error"), tr("Out of memory in aviadd"));
	goto norm_ret;
	}

// fliadd - read a FLI or FLC movie file and add information to wssindex
//	database.

int IndexWindow::fliadd(QString path, QString filename, int /* dummy */)
	{
	char tempbuf[TEMPBUF_SIZE];
	unsigned long basefile;
	int return_code;
	struct filerec *dr_nofile;
	FILE *fp;

#pragma pack(2)
	struct fli_header
		{
		long	size;
		unsigned short magic;
		unsigned short frames;
		unsigned short	width;
		unsigned short	height;
		unsigned short	depth;
		} fli;
	std::stringstream ss;

	if(skip_extract(filename))
		return(SKIPPED);

	basefile = Nofile - 1;
	return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	if(1 != fread(&fli, sizeof(struct fli_header), 1, fp))
		goto corrupt;

	if(fli.magic != 0xaf11U && fli.magic != 0xaf12U)
		goto corrupt;

	dr_nofile = Dr + basefile;
	ss << fli.frames << " frames, " << fli.width << "x" << fli.height << "x" << fli.depth;
	strncpy_s(tempbuf, TEMPBUF_SIZE, ss.str().c_str(), TEMPBUF_SIZE-1);
	if(extra_arch(tempbuf, dr_nofile, Max_extra_length))
		goto no_core;

norm_ret:
	fclose(fp);
	return(return_code);

corrupt:
	return_code = CORRUPT;
	Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not an FLI/FLC movie file.")
			.arg(path, filename));
	goto norm_ret;

no_core:
	return_code = DATABASE_FULL;
	Stop_indexing = true;
	QMessageBox::critical(this, tr("Error"), tr("Out of memory in fliadd"));
	goto norm_ret;
	}
