// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"

// pcxadd - read a .PCX file and add information to wssindex database.
//	The image size and number of colors are converted into extra info
//	of the form WIDTHxHEIGHTxNCOLORS, plus monochrome if all colors in
//	palette have equal RGB components, or grayscale if header indicates
//	grayscale.

int IndexWindow::pcxadd(QString path, QString filename, int /* dummy */)
	{
	struct pcx
		{
		char	manufacturer;	// Should be 0xa
		char	version;			// If 5, there may be a 256 color palette
									//				appended to the file
		char	encoding;		// Should be 1
		char	bitsperpixel;	// Bits per pixel per plane
		short	xmin, ymin;		// Starting pixels
		short	xmax, ymax;		// Ending pixels
		short	hres, vres;		// Resolution of scanning device
		char	palette[16][3];// Palette table
		char	reserved;		// Unused, older docs call this vmode
		char	planes;			// Number of planes, may be zero
		short	bytesline;		// bytes per line of image
		short	paletteinfo;	// 1=color/BW, 2=gray scale, may be 0
		char	filler[58];		// unused
		} pcx_header;

	unsigned long basefile;
	int return_code;
	short height, i, ncolor_bits, ncolors, width;
	struct filerec *dr_nofile;
	bool mono;
	FILE *fp;

	if(skip_extract(filename))
		return(SKIPPED);

	return_code = EXTRACTED_OK;
	basefile = Nofile - 1;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	if(1 != fread(&pcx_header, sizeof(struct pcx), 1, fp)
		|| 0xa != pcx_header.manufacturer)
		{
		goto corrupt;			// this isn't an error, just a warning
		}

	width = pcx_header.xmax - pcx_header.xmin + 1;
	height = pcx_header.ymax - pcx_header.ymin + 1;
	ncolor_bits = pcx_header.bitsperpixel;
	if(pcx_header.planes > 0)
		ncolor_bits *= pcx_header.planes;
	ncolors = static_cast<short>(1 << ncolor_bits);

	dr_nofile = Dr + basefile;
	dr_nofile->ExtraInfo = tr("%1x%2x%3").arg(width).arg(height).arg(num_color_str(ncolor_bits));

	if(2 == pcx_header.paletteinfo)
		dr_nofile->ExtraInfo += tr(", grayscale");
	else
		{
		if(ncolors > 16)
			ncolors = 16;
		mono = true;
		for(i=0; i<ncolors; i++)
			{
			if(pcx_header.palette[i][0] != pcx_header.palette[i][1]
				|| pcx_header.palette[i][1] != pcx_header.palette[i][2])
				mono=false;
			}
		if(mono)
			dr_nofile->ExtraInfo += tr(", monochrome");
		}
	setHasExtra(dr_nofile);

	norm_ret:
		fclose(fp);
		return(return_code);

	corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a PCX image.")
				.arg(path, filename));
		goto norm_ret;
	}
