// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/globalsettings.h"
//#include "OtherWssi/wssiproto.h"
#include "OtherWssi/wsxxxadd.h"
#include "Dialogs/indexwindow.h"
#include "Non_member_functions/customio.h"

#include <io.h>
#include <sstream>
#include <iomanip>

#include <QMessageBox>

// If building old version for running under XP, the strxxx_s functions will not be available
#if QT_VERSION < QT_VERSION_CHECK(5, 8, 0)
	#define strncat_s(w,x,y,z) strncat(w,y,z==_TRUNCATE?strlen(y):z)
	#define strncpy_s(w,x,y,z) strncpy(w,y,x)
#endif

#define WAVE_FORMAT_PCM     1		// from mmsystem.h
#define IBM_FORMAT_MULAW	0x0101
#define IBM_FORMAT_ALAW		0x0102
#define IBM_FORMAT_ADPCM	0x0103

//#define UBIT8(Z) (static_cast<unsigned char>(0x ## Z))

#define TEMPBUF_SIZE 256

// wavadd - read a WAV sound file and add information to wssindex
//	database.

int IndexWindow::wavadd(QString path, QString filename, int /* dummy */)
	{
	int have_pieces, nchunk, return_code;
	unsigned long length, seconds;
	unsigned int tenths;
	struct filerec *dr_nofile;
	FILE *fp;

	struct chunk_header
		{
		long id;
		long len;
		} ch;
	struct fmt_chunk
		{
		unsigned short tag;
		unsigned short nchannels;
		unsigned long samplespersec;
		unsigned long avgbytespersec;
		unsigned short blockalign;
		unsigned short bitspersample;
		} fmt;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	have_pieces = nchunk = 0;
	length = 0;
	for(;;)
		{
// Arbitrary limit on number of chunks to avoid infinite loop processing
//	a non-WAVE file

		if(31 == have_pieces)
			break;
		if(nchunk > 10)
			goto corrupt;
		if(1 != fread(&ch, sizeof(struct chunk_header), 1, fp))
			goto corrupt;
		switch(ch.id)
			{
			case 0x46464952UL:				// "RIFF"
				ch.len = 0;
				have_pieces |= 1;
				break;
			case 0x45564157UL:				// "WAVE"
				have_pieces |= 2;
				ch.len = -4;
				break;
			case 0x20746D66UL:		 		// "fmt "
				have_pieces |= 4;
				if(1 != fread(&fmt, sizeof(struct fmt_chunk), 1, fp))
					goto corrupt;
				ch.len -= sizeof(struct fmt_chunk);
				if(WAVE_FORMAT_PCM == fmt.tag)
					have_pieces |= 16;				// Don't need fact chunk
				break;
			case 0x61746164UL:				// "data"
				have_pieces |= 8;
				length = static_cast<unsigned long>(ch.len);
				break;
			case 0x74636166UL:				// "fact"
				have_pieces |= 16;
				length = getlong(fp, false);
				ch.len -= 4;
				break;
			default:
				break;
			}

		nchunk++;
		fseek(fp, ch.len, SEEK_CUR);
		}

	if(fmt.nchannels > 1)
		length /= fmt.nchannels;
	if(fmt.bitspersample > 0)
		length = (8 * length) / fmt.bitspersample;
	length += fmt.samplespersec/20;		// round up
	seconds = length / fmt.samplespersec;
	tenths = static_cast<unsigned int>(((length - seconds*fmt.samplespersec) * 10) /
						fmt.samplespersec);
	dr_nofile = Dr + basefile;
	switch(fmt.nchannels)
		{
		case 1:
			dr_nofile->ExtraInfo = tr("mono WAVE");
			break;
		case 2:
			dr_nofile->ExtraInfo = tr("stereo WAVE");
			break;
		default:
			dr_nofile->ExtraInfo = tr("%1 channel WAVE").arg(fmt.nchannels);
			break;
		}
	dr_nofile->ExtraInfo += tr(", %1 Hz, %2.%3 seconds").arg(fmt.samplespersec).arg(seconds).arg(tenths);
	setHasExtra(dr_nofile);

norm_ret:
		fclose(fp);
		return(return_code);

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a WAVE sound file.")
				.arg(path, filename));
		goto norm_ret;
	}

// auadd - read an AU sound file and add information to wssindex
//	database.

int IndexWindow::auadd(QString path, QString filename, int /* dummy */)
	{
	char tempbuf[TEMPBUF_SIZE];
	int comment_length, i, return_code;
	unsigned int tenths;
	unsigned long seconds;
	struct filerec *dr_nofile;
	FILE *fp;
	struct SNDSoundStruct
		{
		long magic;					// magic number SND_MAGIC
		long dataLocation;		// offset or pointer to the data
		long dataSize;				// number of bytes of data
		long dataFormat;			// the data format code
		long samplingRate;		// the sampling rate
		long channelCount;		// the number of channels
		} au;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	if(1 != fread(&au, sizeof(struct SNDSoundStruct), 1, fp))
		goto corrupt;

	for(i=0; i<6; i++)
		swap4(i + &(au.magic));

	if(0x2e736e64L != au.magic)				// .snd
		goto corrupt;
	if(au.channelCount > 1)
		au.dataSize /= au.channelCount;
	au.dataSize += au.samplingRate/20;		// round up
	seconds = static_cast<unsigned long>(au.dataSize / au.samplingRate);
	tenths = static_cast<unsigned int>(((au.dataSize - static_cast<long>(seconds)*au.samplingRate) * 10) /
						au.samplingRate);
	dr_nofile = Dr + basefile;
	switch(static_cast<int>(au.channelCount))
		{
		case 1:
			dr_nofile->ExtraInfo = tr("mono AU");
			break;
		case 2:
			dr_nofile->ExtraInfo = tr("stereo AU");
			break;
		default:
			dr_nofile->ExtraInfo = tr("%1 channel AU").arg(au.channelCount);
			break;
		}
	dr_nofile->ExtraInfo += tr(", %1 Hz, %2.%3 seconds").arg(au.samplingRate).arg(seconds).arg(tenths);
	setHasExtra(dr_nofile);

	if( (comment_length = static_cast<int>(au.dataLocation) - 24) > 0)
		{
		comment_length = qMin(comment_length, static_cast<int>(Max_comment_length));
		if(1 != fread(tempbuf, static_cast<size_t>(comment_length), 1, fp))
			goto corrupt;
		if(com_arch(tempbuf, dr_nofile, comment_length))
			goto no_core;
		}

norm_ret:
	fclose(fp);
	return(return_code);

corrupt:
	return_code = CORRUPT;
	Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not an AU sound file.")
				.arg(path, filename));
	goto norm_ret;

no_core:
	return_code = DATABASE_FULL;
	Stop_indexing = true;
	QMessageBox::critical(this, tr("Error"), tr("Out of memory in auadd"));
	goto norm_ret;
	}

// modadd - read a MOD sound file and add information to wssindex
//	database.

int IndexWindow::modadd(QString path, QString filename, int /* dummy */)
	{
	char tempbuf[TEMPBUF_SIZE];
	int return_code;
	struct filerec *dr_nofile;
	FILE *fp;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	return_code = SKIPPED;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

// Read 20 bytes of song name and 22 bytes of sample name

	if(1 != fread(tempbuf+128, 42, 1, fp))
		goto corrupt;

// Windows has .mod files which are executable, first two bytes
//	are "MZ".  Silently skip these.

	if(0 == strncmp(tempbuf+128, "MZ", 2))
		goto norm_ret;

	return_code = EXTRACTED_OK;
	memset(tempbuf, 0, 44);
	strncpy(tempbuf, tempbuf+128, 20);
	strncat_s(tempbuf, TEMPBUF_SIZE, " ", _TRUNCATE);
	strncat_s(tempbuf, TEMPBUF_SIZE, tempbuf+148, 22);

	dr_nofile = Dr + basefile;
	if(com_arch(tempbuf, dr_nofile, 43))
		goto no_core;

norm_ret:
	fclose(fp);
	return(return_code);

corrupt:
	return_code = CORRUPT;
	Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a MOD sound file.")
			.arg(path, filename));
	goto norm_ret;

no_core:
	return_code = DATABASE_FULL;
	Stop_indexing = true;
	QMessageBox::critical(this, tr("Error"), tr("Out of memory in modadd"));
	goto norm_ret;
	}

// vocadd - read a VOC sound file and add information to wssindex
//	database.

int IndexWindow::vocadd(QString path, QString filename, int /* dummy */)
	{
	char tempbuf[TEMPBUF_SIZE];
	int have_pieces, nchannels, nchunk, nread;
	unsigned long data_length, ext_data, sample_rate, seconds = 0;
	unsigned int tenths = 0;
	struct filerec *dr_nofile;
	FILE *fp;
	unsigned short length_use;
	long header, length, offset;
	char id;

	if(skip_extract(filename))
		return(SKIPPED);

	unsigned long basefile = Nofile - 1;
	int return_code = EXTRACTED_OK;

	if(nullptr == (fp = arc_open(path, filename)))
		return(CANT_OPEN);

	if(1 != fread(tempbuf, 20, 1, fp))
		goto corrupt;
	if(0 != strncmp(tempbuf, "Creative Voice File", 19))
		goto corrupt;
	dr_nofile = Dr + basefile;
	offset = static_cast<long>(fgetWord(fp)) - 22;
	have_pieces = nchunk = 0;
	nchannels = 1;
	data_length = sample_rate = 0;
	for(;;)
		{
// Arbitrary limit on number of chunks to avoid infinite loop processing
//	a non-VOC file
		if(have_pieces > 2)
			break;
		if(nchunk > 20)
			goto corrupt;
		fseek(fp, offset, SEEK_CUR);
// Terminator is only one byte
		nread = static_cast<int>(fread(&header, 1, 4, fp));
		id = static_cast<char>(header & 0xff);
		length = header >> 8;
		if(0 == nread)
			goto corrupt;
		if(0 == id)				// terminator
			{
//			have_pieces |= 15;
			break;
			}
		if(4 != nread)
			goto corrupt;
		switch(id)
			{
			case 1:		// sound data
				have_pieces |= 1;
				sample_rate = static_cast<unsigned long>(256 - fgetc(fp));
				if(sample_rate > 0)
					sample_rate = 1000000UL / sample_rate;
				data_length = static_cast<unsigned long>(length - 2);
				offset = length - 1;
				break;
			case 5:		// ascii (comment/title)
				if(hasComment(dr_nofile))
					{
					offset = length;	// ignore second comment
					break;
					}
				length_use = static_cast<unsigned short>(qMin(length, static_cast<long>(Max_comment_length)));
				if(1 != fread(tempbuf, length_use, 1, fp))
					goto corrupt;

				if(com_arch(tempbuf, dr_nofile, length_use))
					goto no_core;
				offset = length - length_use;
				break;
			case 8:		// extended (may not be present)
				have_pieces |= 2;
				if(1 != fread(&ext_data, 4, 1, fp))
					goto corrupt;
				nchannels = 0 == (ext_data & 0x80000000UL) ? 1 : 2;
				offset = length - 4;
				break;
			default:
				offset = length;
				break;
			}
		nchunk++;
		}
	if(sample_rate > 0)
		{
		data_length += sample_rate/20;		// round up
		seconds = data_length / sample_rate;
		tenths = static_cast<unsigned int>(((data_length - seconds*sample_rate) * 10) / sample_rate);
		}
	switch(nchannels)
		{
		case 1:
			dr_nofile->ExtraInfo = tr("mono VOC");
			break;
		case 2:
			dr_nofile->ExtraInfo = tr("stereo VOC");
			break;
		default:
			break;
		}
	dr_nofile->ExtraInfo += tr(", %1 Hz, %2.%3 seconds").arg(sample_rate).arg(seconds).arg(tenths);
	setHasExtra(dr_nofile);

norm_ret:
		fclose(fp);
		return(return_code);

corrupt:
		return_code = CORRUPT;
		Q_EMIT signal_logMessage(tr("%1%2 is corrupted or not a VOC sound file.")
				.arg(path, filename));
		goto norm_ret;

no_core:
		return_code = DATABASE_FULL;
		Stop_indexing = true;
		QMessageBox::critical(this, tr("Error"), tr("Out of memory in vocadd"));
		goto norm_ret;
	}
