// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#ifndef EXTRACTINFO_H
#define EXTRACTINFO_H

#include <QString>
#include <QRegExp>

#ifndef EXTERN
	#define EXTERN extern
#endif

#define TYPE_ARCHIVE				0
#define TYPE_GRAPHICS			1
#define TYPE_SELF_EXTRACTING	2
#define TYPE_FONT					3
#define TYPE_SOUND				4
#define TYPE_MOVIE				5
#define TYPE_ALL					6

// The order here doesn't matter, new values can be added anywhere convenient
enum ExtractRoutine
	{
//Archives
	zip7Extract, arcExtract, arjExtract, arkExtract, cabExtract,
	hapExtract, hypExtract, lbrExtract, lhaExtract, lzsExtract,
	pakExtract, pkaExtract, rqqExtract, sdnExtract, zipExtract,
	zooExtract,
//Fonts
	gsfExtract, pfaExtract, pfbExtract, ttfExtract,
// Images
	bmpExtract, gifExtract, jpgExtract, pcxExtract, pngExtract,
	tgaExtract, tifExtract,
// Video
	aviExtract, flcExtract, fliExtract, flvExtract, mpgExtract,
	mkvExtract, 	mp4Extract, ogmExtract, qtExtract, swfExtract,
	wmvExtract,
// Audio
	auExtract, flacExtract, m4aExtract, modExtract, mp3Extract,
	vocExtract, wavExtract,
// Self-extracting exe files
	arcExtractExe, arjExtractExe, lhaExtractExe, zipExtractExe,
// Leave this one at end of list so NLIBTYPE gets correct value
	endlistExtract
	};

// Compiling main.cpp will give an error if NLIBTYPE is too small
// This is a trick so NLIBTYPE get the right value automatically when
// additional file types are added to the extraction list
#define NLIBTYPE	endlistExtract

#define dontExtract					0
#define extractMembers				1
#define extractInfo					1
#define extractArcComment			2
#define extractMembersOrComment	3
#define extractMemberComments		4
#define extractAll					7

EXTERN struct extrct
	{
	QString description;		// Human-readable file extension(s), also registry key for saved extraction option
	QRegExp nameRegexp;		// Regular expression matching file extension
	QString tooltip;			// Optional tooltip text
	ExtractRoutine routine;	// Extraction routine from enum list
	int flag;					// Default extraction options for file type
	int valid_flag_values;	// Allowed extraction options for file type
	int type;					// TYPE_ARCHIVE, GRAPHICS,...
	} Extract[NLIBTYPE]
// Following will be true if EXTERN is defined but empty
#if defined(DO_INITIALIZATIONS)
		= {
// The order here doesn't matter, for convenience, it matches the order of the ExtractRoutine enum
//Archives
		{"7-zip", QRegExp("^7z$", Qt::CaseInsensitive),
			QString(), zip7Extract,
			extractMembers, extractMembers, TYPE_ARCHIVE},
		{"ARC", QRegExp("^ARC$", Qt::CaseInsensitive),
			QT_TR_NOOP("ARC does not support comments, but PKArc and PAK extended the format"), arcExtract,
			extractMembersOrComment, extractMembersOrComment, TYPE_ARCHIVE},
		{"ARJ", QRegExp("^ARJ$|^A\\d\\d$", Qt::CaseInsensitive),
			QString(), arjExtract,
			extractMembersOrComment, extractMembersOrComment, TYPE_ARCHIVE},
		{"ARK", QRegExp("^ARK$", Qt::CaseInsensitive),
			QT_TR_NOOP("ARC format written by PKARC"), arkExtract,
			extractMembersOrComment, extractMembersOrComment, TYPE_ARCHIVE},
		{"CAB", QRegExp("^CAB$", Qt::CaseInsensitive),
			QString(), cabExtract,
			extractMembers, extractMembers, TYPE_ARCHIVE},
		{"HAP", QRegExp("^HAP$", Qt::CaseInsensitive),
			QString(), hapExtract,
			extractMembers, extractMembers, TYPE_ARCHIVE},
		{"HYP", QRegExp("^HYP$", Qt::CaseInsensitive),
			QString(), hypExtract,
			extractMembers, extractMembers, TYPE_ARCHIVE},
		{"LBR", QRegExp("^LBR$", Qt::CaseInsensitive),
			QString(), lbrExtract,
			extractMembers, extractMembers, TYPE_ARCHIVE},
		{"LHA, LZH", QRegExp("^LHA$|^LZH$", Qt::CaseInsensitive),
			QString(), lhaExtract,
			extractMembers, extractMembers, TYPE_ARCHIVE},
		{"LZS", QRegExp("^LZS$", Qt::CaseInsensitive),
			QString(), lzsExtract,
			extractMembers, extractMembers, TYPE_ARCHIVE,},
		{"PAK", QRegExp("^PAK$", Qt::CaseInsensitive),
			QT_TR_NOOP("ARC file variant written by PAK"), pakExtract,
			extractMembersOrComment, extractMembersOrComment, TYPE_ARCHIVE},
		{"PKA", QRegExp("^PKA$", Qt::CaseInsensitive),
			QT_TR_NOOP("ARC file variant written by PKArc"), pkaExtract,
			extractMembersOrComment, extractMembersOrComment, TYPE_ARCHIVE},
		{"RAR", QRegExp("^RAR$|^R\\d\\d$", Qt::CaseInsensitive),
			QString(), rqqExtract,
			extractMembersOrComment, extractMembersOrComment, TYPE_ARCHIVE},
		{"SDN", QRegExp("^SDN$", Qt::CaseInsensitive),
			QT_TR_NOOP("An old ARC variant from Fidonet Software Distribution Network"), sdnExtract,
			extractMembersOrComment, extractMembersOrComment, TYPE_ARCHIVE},
		{"ZIP", QRegExp("^ZIP$", Qt::CaseInsensitive),
			QString(), zipExtract,
			extractAll, extractAll, TYPE_ARCHIVE,},
		{"ZOO", QRegExp("^ZOO$", Qt::CaseInsensitive),
			QString(), zooExtract,
			extractAll, extractAll, TYPE_ARCHIVE,},

// Fonts
		{"GSF", QRegExp("^GSF$", Qt::CaseInsensitive),
			QT_TR_NOOP("Ghostscript font"), gsfExtract,
			extractAll, extractAll, TYPE_FONT,},
		{"PFA", QRegExp("^PFA$", Qt::CaseInsensitive),
			QT_TR_NOOP("Type A postscript font"), pfaExtract,
			extractAll, extractAll, TYPE_FONT},
		{"PFB", QRegExp("^PFB$", Qt::CaseInsensitive),
			QT_TR_NOOP("Type B postscript font"), pfbExtract,
			extractAll, extractAll, TYPE_FONT},
		{"TTF", QRegExp("^TTF$", Qt::CaseInsensitive),
			QT_TR_NOOP("Windows True Type font"), ttfExtract,
			extractAll, extractAll, TYPE_FONT},

// Images
		{"BMP", QRegExp("^BMP$", Qt::CaseInsensitive),
			QString(), bmpExtract,
			extractAll, extractAll, TYPE_GRAPHICS},
		{"GIF", QRegExp("^GIF$", Qt::CaseInsensitive),
			QString(), gifExtract,
			extractAll, extractAll, TYPE_GRAPHICS},
		{"JPG, JPEG", QRegExp("^JPE?G$", Qt::CaseInsensitive),
			QString(), jpgExtract,
			extractAll, extractAll, TYPE_GRAPHICS},
		{"PCX", QRegExp("^PCX$", Qt::CaseInsensitive),
			QString(), pcxExtract,
			extractAll, extractAll, TYPE_GRAPHICS},
		{"PNG", QRegExp("^PNG$", Qt::CaseInsensitive),
			QString(), pngExtract,
			extractAll, extractAll, TYPE_GRAPHICS},
		{"TGA", QRegExp("^TGA$", Qt::CaseInsensitive),
			QString(), tgaExtract,
			extractAll, extractAll, TYPE_GRAPHICS},
		{"TIF, TIFF", QRegExp("^TIFF?$", Qt::CaseInsensitive),
			QString(), tifExtract,
			extractAll, extractAll, TYPE_GRAPHICS},

// Video
		{"AVI", QRegExp("^AVI$", Qt::CaseInsensitive),
			QString(), aviExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"FLC", QRegExp("^FLC$", Qt::CaseInsensitive),
			QString(), flcExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"FLI", QRegExp("^FLI$", Qt::CaseInsensitive),
			QString(), fliExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"FLV", QRegExp("^FLV$", Qt::CaseInsensitive),
			QString(), flvExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"MPG, MPEG", QRegExp("^MPE?G$", Qt::CaseInsensitive),
			QString(), mpgExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"MKV", QRegExp("^MKV$", Qt::CaseInsensitive),
			QT_TR_NOOP("Matroska"), mkvExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"MP4", QRegExp("^MP4$", Qt::CaseInsensitive),
			QT_TR_NOOP("MPEG-4"), mp4Extract,
			extractAll, extractAll, TYPE_MOVIE},
		{"MOV", QRegExp("^MOV$", Qt::CaseInsensitive),
			QT_TR_NOOP("Quicktime"), qtExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"OGM", QRegExp("^OGM$", Qt::CaseInsensitive),
			QT_TR_NOOP("Ogg video"), ogmExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"SWF", QRegExp("^SWF$", Qt::CaseInsensitive),
			QT_TR_NOOP("Shockwave Flash"), swfExtract,
			extractAll, extractAll, TYPE_MOVIE},
		{"WMV", QRegExp("^WMV$", Qt::CaseInsensitive),
			QT_TR_NOOP("Windows Media"), wmvExtract,
			extractAll, extractAll, TYPE_MOVIE},

// Audio
		{"AU", QRegExp("^AU$", Qt::CaseInsensitive),
			QString(), auExtract,
			extractAll, extractAll, TYPE_SOUND},
		{"FLAC", QRegExp("^FLAC$", Qt::CaseInsensitive),
			QT_TR_NOOP("Lossless audio"), flacExtract,
			extractAll, extractAll, TYPE_SOUND},
		{"M4A", QRegExp("^M4A$", Qt::CaseInsensitive),
			QT_TR_NOOP("iTunes"), m4aExtract,
			extractAll, extractAll, TYPE_SOUND},
		{"MOD", QRegExp("^MOD$", Qt::CaseInsensitive),
			QString(), modExtract,
			extractAll, extractAll, TYPE_SOUND},
		{"MP3", QRegExp("^MP3$", Qt::CaseInsensitive),
			QString(), mp3Extract,
			extractAll, extractAll, TYPE_SOUND},
		{"VOC", QRegExp("^VOC$", Qt::CaseInsensitive),
			QString(), vocExtract,
			extractAll, extractAll, TYPE_SOUND},
		{"WAV", QRegExp("^WAV$", Qt::CaseInsensitive),
			QString(), wavExtract,
			extractAll, extractAll, TYPE_SOUND},

// Self-extracting exe files
		{"ARC", QRegExp("^EXE$", Qt::CaseInsensitive),	// PAK adds self-extracting option to ARC
			QString(), arcExtractExe,
			dontExtract, extractMembersOrComment, TYPE_SELF_EXTRACTING},
		{"ARJ", QRegExp("^EXE$", Qt::CaseInsensitive),
			QString(), arjExtractExe,
			dontExtract, extractMembersOrComment, TYPE_SELF_EXTRACTING},
		{"LHA", QRegExp("^EXE$", Qt::CaseInsensitive),
			QString(), lhaExtractExe,
			dontExtract, extractMembers, TYPE_SELF_EXTRACTING},
		{"ZIP", QRegExp("^EXE$", Qt::CaseInsensitive),
			QString(), zipExtractExe,
			dontExtract, extractAll, TYPE_SELF_EXTRACTING}
		}
#endif
		;

#endif // EXTRACTINFO_H
