// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "ModelView/filetablemodel.h"
#include "OtherWssi/tablelayout.h"
#include "Dialogs/mainwindow.h"
#include "ui_mainwindow.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include <QAbstractItemModel>
#include <QtDebug>
#include <QItemSelectionModel>
#include <QToolTip>
#include <QCursor>

fileTableModel::fileTableModel(QObject *parent) : QAbstractTableModel(parent)
	{
	fileModelValid = false;
	}

void fileTableModel::signalLayoutChanged()
	{
	Q_EMIT QAbstractItemModel::layoutChanged();
	}

int fileTableModel::rowCount(const QModelIndex & /*parent*/) const
	{
	return fileModelValid ? modelRows : 0;
	}

void fileTableModel::setFileModelValidity(bool isvalid)
	{
	fileModelValid = isvalid;
	}

int fileTableModel::columnCount(const QModelIndex & /*parent*/) const
	{
	return fileModelValid ? modelCols : 0;
	}

void fileTableModel::setFields()
	{
	memset(fieldNumber, -1, 4*NFIELDS*sizeof(int));
	for(int i=0; i<NFIELDS; i++)
		{
		if(fieldlayout[i].enabled)
			fieldNumber[fieldlayout[i].row][fieldlayout[i].column] = i;
		}
	}

void fileTableModel::setSizeAndValidity(int rows, int cols, unsigned long file_start)
	{
	beginResetModel();
	modelRows = rows;
	modelCols = cols;
	firstFile = file_start;
	fileModelValid = true;
	endResetModel();
	}

QVariant fileTableModel::data(const QModelIndex &index, int role) const
	{
// index.row() < 0 happens after clicking header to sort and then typing ^C.
// No other keystroke seems to trigger this.
	if(!fileModelValid || index.row() < 0)
		return QVariant();

	unsigned long arcbase;
	bool inarc;
	unsigned long filenum = firstFile + static_cast<unsigned long>(index.row()/rowsPerItem);
	int smallRow = index.row() % rowsPerItem;
	int col = index.column();

	QString blank_indent;
	blank_indent.fill(QLatin1Char(' '), Indent_members);

	unsigned long k = Sortindx[filenum];
	struct filerec *dr_k = Dr + k;
	unsigned long volno = dr_k->Volno;
	struct diskrec *vrec_volno = Vrec + volno;
	if(inArc(dr_k))
		{
		inarc = true;
		arcbase = dr_k->Subdirno;
		}
	else
		{
		inarc = false;
		arcbase = k;
		}
	struct filerec *dr_arcbase = Dr + arcbase;

// If this is a cell hidden by a span or otherwise unused, fieldNo will be -1
// and only background color needs to be set
	int fieldNo = fieldNumber[smallRow][col];

	switch(role)
		{
		case Qt::DisplayRole:
			if(-1 == fieldNo)
				break;
			if(fieldlayout[fieldNo].width > 1 || fieldlayout[fieldNo].height > 1)
				{
				int thisRow = index.row()-smallRow+fieldlayout[fieldNo].row;
				int thisCol = fieldlayout[fieldNo].column;
				mainWindowPtr->ui->tableView->setSpan(thisRow, thisCol,
					fieldlayout[fieldNo].height, fieldlayout[fieldNo].width);
				}
			Q_FALLTHROUGH();
		case Qt::EditRole:
			if(-1 == fieldNo)
				break;
			switch(fieldNo)
				{
				case FIELD_LFN:
					if(isDotID(dr_k))
						{
						if(0 == dr_k->Subdirno)
							return(tr("<Volume>"));
						return(tr("<Subdir>"));
						}
					if(Indent_members>0 && inArc(dr_k))
						return(QString(blank_indent + dr_k->Lfn));
					return(dr_k->Lfn);
				case FIELD_EXT:
					if(dr_k->ExtOffset > 0 && !isDotID(dr_k))
						return(dr_k->Lfn.mid(dr_k->ExtOffset));
					break;
				case FIELD_VOL:
					return(vrec_volno->VolName);
				case FIELD_SUBDIR:
					if(!isDotID(dr_k) || dr_k->Subdirno > 0)
						return(QDir::toNativeSeparators(SubDir[dr_arcbase->Subdirno]));
					break;
				case FIELD_ARCHIVE:
					if(inarc)
						return(dr_arcbase->Lfn);
					break;
				case FIELD_SIZE:
					if(!isDotID(dr_k))
						return(formattedsize(dr_k->Dsize, Filesize_unit));
					break;
				case FIELD_TIMESTAMP:
					if(dr_k->Ddate.isValid() && dr_k->Ddate.year() > 1600)
						{
						if(DateOnly)
							return(dr_k->Ddate.toString(ShortDateFormat));
						else
							{
							return(QString(dr_k->Ddate.toString(ShortDateFormat)
										  + " " + dr_k->Dtime.toString()));
							}
						}
					break;
				case FIELD_CATEGORY:
					if(categoryEdited(dr_k) && dr_k->Category.isEmpty())
						return(tr("<Deleted>"));
					return(dr_k->Category);
				case FIELD_COMMENT:
					if(commentEdited(dr_k) && dr_k->Comment.isEmpty())
						return(tr("<Deleted>"));
					return(dr_k->Comment);
				case FIELD_EXTRA_INFO:
					if(extraEdited(dr_k) && dr_k->ExtraInfo.isEmpty())
						return(tr("<Deleted>"));
					return(dr_k->ExtraInfo);
				default:
				case -1:
					break;
				}
			return "";

		case Qt::BackgroundRole:
			{
			QColor rowcolor;
			if(FileUseAltColors)
				rowcolor = FileRowColor[filenum&1];
			else
				rowcolor = FileRowColor[0];
			return rowcolor;
			}

		case Qt::ForegroundRole:
			if(-1 == fieldNo)
				break;
			{
			QColor txtcolor = Qt::black;
			switch(fieldNumber[smallRow][col])
				{
				case FIELD_VOL:
					if(volumeEdited(dr_k))
						txtcolor = UpdatedColor;
					break;
				case FIELD_CATEGORY:
					if(categoryEdited(dr_k))
						txtcolor = UpdatedColor;
					break;
				case FIELD_COMMENT:
					if(commentEdited(dr_k))
						txtcolor = UpdatedColor;
					break;
				case FIELD_EXTRA_INFO:
					if(extraEdited(dr_k))
						txtcolor = UpdatedColor;
					break;
				}
			return txtcolor;
			}

		case Qt::TextAlignmentRole:
			if(-1 == fieldNo)
				break;
			return(fieldlayout[fieldNumber[smallRow][col]].alignment);

		case Qt::FontRole:
			return *Font_filelist;

		case Qt::SizeHintRole:
			break;

		case Qt::ToolTipRole:
			if(index.row() != mainWindowPtr->fileSelection->currentIndex().row()
					|| index.column() != mainWindowPtr->fileSelection->currentIndex().column())
				break;
			switch(fieldNo)
				{
				case FIELD_VOL:
				case FIELD_CATEGORY:
				case FIELD_COMMENT:
				case FIELD_EXTRA_INFO:
					return(tr("F2 or double-click to edit"));

				default:
					break;
				}
			break;

		default:
			break;
		}
	return QVariant();
	}

QVariant fileTableModel::headerData(int section, Qt::Orientation orientation, int role) const
	{
	if (Qt::DisplayRole != role || Qt::Horizontal != orientation)
		return QVariant();
	return theader[section];
	}

bool fileTableModel::setHeaderData(int section, Qt::Orientation orientation,
											  const QVariant &value, int role)
	{
	if (Qt::DisplayRole != role || Qt::Horizontal != orientation)
		return false;
	theader[section] = value.toString();
	Q_EMIT headerDataChanged(orientation, section, section);
	return true;
	}

void fileTableModel::refreshData()
	{
	Q_EMIT dataChanged(QModelIndex(), QModelIndex());
	}

// Category, comment, extra info and volume cells are editable, others are not
Qt::ItemFlags fileTableModel::flags(const QModelIndex &index) const
	{
	int smallRow = index.row() % rowsPerItem;
	int col = index.column();
	switch(fieldNumber[smallRow][col])
		{
		case FIELD_VOL:
		case FIELD_CATEGORY:
		case FIELD_COMMENT:
		case FIELD_EXTRA_INFO:
			return(Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable);

		default:
			break;
		}

	return(Qt::ItemIsEnabled | Qt::ItemIsSelectable);
	}

bool fileTableModel::setData(const QModelIndex & index, const QVariant & value, int role)
	{
	if (role == Qt::EditRole)
		{
	  Q_EMIT signal_fileEditCompleted(index, value.toString());
		}
	return true;
	}
