// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "OtherWssi/wssiproto.h"
#include "OtherWssi/globalsettings.h"
#include <QMessageBox>
#include <QApplication>

void mergeworker(unsigned long *array, unsigned long lo, unsigned long hi,
	unsigned long *workspace);
void merge(unsigned long *array, unsigned long lo, unsigned long m, unsigned long hi,
	unsigned long *workspace);
void binaryInsertionSort(unsigned long *array, unsigned long n);

static int (*cmpfunc)(const unsigned long, const unsigned long);

// Merge sort array, dimension n.  cmpf is the comparison function.
// This is a fairly straightforward implementation.  A good description
// of the algorithm can be found at
// https://www.inf.fh-flensburg.de/lang/algorithmen/sortieren/merge/mergen.htm

// Main routine - allocates temporary space needed by merge sort and calls
// mergeworker to do the sort.
void mergesort(unsigned long *array, unsigned long n,
					int (*cmpf)(const unsigned long, const unsigned long))
	{
	unsigned long *workspace;
	cmpfunc = cmpf;
	if(n<=1)
		return;
	try {
		workspace = new unsigned long[(n+1)/2];
		}
	catch (std::bad_alloc&)
		{
		QMessageBox::critical(nullptr, QT_TRANSLATE_NOOP("mergesort", "Error"),
			QT_TRANSLATE_NOOP("mergesort", "Insufficient memory to sort database"));
		return;
		}

	unsigned long lo = 0;
	unsigned long hi = n-1;
	mergeworker(array, lo, hi, workspace);
	delete [] workspace;
	}

// Sort is recursive.  Split the array in half and call yourself twice
// to sort the two halves, then merge the two sorted lists.

// This was for special-casing when the pieces to be sorted get small
// which didn't make significant changes to sort times.
// If used, a trailing backslash would be added to the #define

//#define CMP_SWAP(x,y) if(cmpfunc(*(x), *(y)) > 0)
//	{
//	unsigned long tmp = *(y);
//	*(y) = *(x);
//	*(x) = tmp;
//	}

void mergeworker(unsigned long *array, unsigned long lo, unsigned long hi,
	unsigned long *workspace)
	{
	if(hi <= lo)
		return;

	unsigned long m = (lo+hi)/2;
	mergeworker(array, lo, m, workspace);
	mergeworker(array, m+1, hi, workspace);
	merge(array, lo, m, hi, workspace);
	}

void merge(unsigned long *array, unsigned long lo, unsigned long m,
	unsigned long hi, unsigned long *workspace)
	{
	unsigned long i, j, k;

	i = 0;
	j = lo;
// copy first half of array a to auxiliary array workspace
//	while(j<=m)
//		workspace[i++] = array[j++];
	if(j<=m)
		{
		memcpy(workspace+i, array+j, (m+1-j)*sizeof(unsigned long));
		j = m + 1;
		}

	i=0;
	k=lo;
// copy back next-greatest element at each time
	while (k<j && j<=hi)
		if(cmpfunc(workspace[i], array[j]) <= 0)
			array[k++] = workspace[i++];
		else
			array[k++] = array[j++];

// copy back remaining elements of first half (if any)
//	while (k<j)
//		array[k++] = workspace[i++];
	if(k<j)
		{
		memcpy(array+k, workspace+i, (j-k)*sizeof(unsigned long));
		}
	}
