// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include <stdio.h>
#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include "Non_member_functions/customio.h"
#include "OtherWssi/globalsettings.h"
#include "OtherWssi/wssiproto.h"
#include "ModelView/filetablemodel.h"

#include <QDebug>

#define IO_BUF_SIZE 32768

// getwss - read an existing wssindex database into memory.
// Return true if read was successful, false if not.

#define READ_16_BITS	true
#define READ_32_BITS	false

bool MainWindow::getwss()
	{
	FILE *fp;						// file pointer for database being read
	struct filerec	*dr_fix,		// pointer to archive file containing file
										//	being patched
						*dr_temp;	// pointer to current file record
	struct diskrec	*vrec_i;		// pointer to current disk record

	unsigned char 	tempc;		// temp var reads 'C' or blank for com/cat

	bool				consistent = true,	// set false if structural error found
						read_lfn = false,		// true if reading post 6.01 database with
													//	diskbits flags and long file names
						vols_ok = true;		// set false if file with invalid volume
													//	number found
// warn_truncate is old code, wgets silently truncates
//						warn_truncate = false;	// set true if warning needed about
														//	long comment truncation

	bool				disk_size_64_bit,
						vol_number_32_bit,
						VolName_length_fixed,
						total_subdir_number_32_bit,
						subdirs_one_disk_32_bit,
						total_file_count_32_bit,
						file_count_32_bit,
						file_size_64_bit,
						file_size_before_date,
						short_fn_ext_present,
						diskbits_one_byte,
						old_format_no_categories,
						bootable_flag_present,
						msdos_time_format,
						unicode_strings;

	unsigned int diskstart,			// first disk no. to read, 0 unless merging
						filestart,		// first file no. to read, 0 unless merging
						flag_vol,		// last volume flagged for reindexing
						last_good_vol,	// volume number of last good file, with high
											//	bit cleared
						local_subdirno,// subdir number where inconsistency found
						local_volno,	// volume number where inconsistency found
						newdiskno,		// no. disks there will be if read succeeds
						newnofile,		// no. files there will be if read succeeds
						newsubno,		// no. subs there will be if read succeeds
						substart;		// first sub no. to read, 0 unless merging

	long				nfread,			// number of last file read or -1
						nvread,			// number of last volume read or -1
						nsread;			// number of last subdir read or -1

	char 				*iobuff;			// pointer to buffer for reading database

	bool				merging=false;	// true if merging databases

	unsigned long	prevVolNo;
#define STRINGBUFFER_SIZE 4096
	char				stringbuffer[STRINGBUFFER_SIZE];

	diskstart = filestart = substart = 0;
	if(Diskno>0)
		{
		QMessageBox msgBox;
		msgBox.setText(tr("Merge with in-memory database?"));
		if(Changed)
			msgBox.setInformativeText(tr("<b>The current database has changes which have not been saved.</b>"));
		QPushButton *mergeButton = msgBox.addButton(tr("Merge databases"), QMessageBox::ActionRole);
		msgBox.addButton(tr("Discard current database"), QMessageBox::ActionRole);
		QPushButton *cancelButton  = msgBox.addButton(QMessageBox::Cancel);
		msgBox.setDefaultButton(mergeButton);
		msgBox.exec();
		if(static_cast<QPushButton *>(msgBox.clickedButton()) == cancelButton)
			return false;
		if(static_cast<QPushButton *>(msgBox.clickedButton()) == mergeButton)
			merging = true;
		}

	if(nullptr == (fp = _wfopen(database_name.toStdWString().data(), L"rb")))
		{
		QMessageBox::critical(this, tr("Error"), tr("Can't open input file %1").arg(database_name));
		return false;
		}

	Ftm->setFileModelValidity(false);
	Vtm->setVolModelValidity(false);
	if(merging)
		{
		diskstart = Diskno;
		filestart = Nofile;
		substart = Subno;
		}
	else
		{
		freeall();
		fileBalance->colWidthsValid = false;
		}

	if(nullptr == (iobuff = static_cast<char *>(malloc(IO_BUF_SIZE))))
		goto nocore;

	setvbuf(fp, iobuff, _IOFBF, IO_BUF_SIZE);

	nvread = nsread = nfread = -1;
	if(!wgets(stringbuffer, 9, fp, false))
		goto readerr;
// The Wssindex database format has changed over time.
// These flags define the format being read.
// The defaults are for Wssindex 8.0, the first version using the Qt Toolkit.
// Wssindex 9 uses the same format as 8.
// Some formats may never have been released.
	bootable_flag_present = false;
	disk_size_64_bit = true;
	diskbits_one_byte = false;
	file_count_32_bit = true;
	file_size_64_bit = true;
	file_size_before_date = true;
	msdos_time_format = false;
	old_format_no_categories = false;
	read_lfn = true;
	short_fn_ext_present = false;
	subdirs_one_disk_32_bit = true;
	total_file_count_32_bit = true;
	total_subdir_number_32_bit = true;
	unicode_strings = true;
	VolName_length_fixed = false;
	vol_number_32_bit = true;
	if(0 == strcmp(stringbuffer, "WSSI-Qt"))
		{
		}
	else if(0 == strcmp(stringbuffer, "WSEXTEND"))
		{
		bootable_flag_present = true;
		diskbits_one_byte = true;
		disk_size_64_bit = false;
		file_size_64_bit = false;
		file_size_before_date = false;
		msdos_time_format = true;
		read_lfn = false;
		short_fn_ext_present = true;
		subdirs_one_disk_32_bit = false;
		unicode_strings = false;
		VolName_length_fixed = true;
		}
	else if(0 == strcmp(stringbuffer, "WSEXT-L"))
		{
		bootable_flag_present = true;
		diskbits_one_byte = true;
		disk_size_64_bit = false;
		file_count_32_bit = true;
		file_size_64_bit = false;
		file_size_before_date = false;
		msdos_time_format = true;
		read_lfn = true;
		short_fn_ext_present = true;
		subdirs_one_disk_32_bit = false;
		total_file_count_32_bit = true;
		total_subdir_number_32_bit = true;
		unicode_strings = false;
		VolName_length_fixed = true;
		vol_number_32_bit = true;
		}
	else if(0 == strcmp(stringbuffer, "WSSI-L"))
		{
		bootable_flag_present = true;
		diskbits_one_byte = true;
		disk_size_64_bit = false;
		file_count_32_bit = false;
		file_size_64_bit = false;
		file_size_before_date = false;
		msdos_time_format = true;
		read_lfn = true;
		short_fn_ext_present = true;
		subdirs_one_disk_32_bit = false;
		total_file_count_32_bit = false;
		total_subdir_number_32_bit = false;
		unicode_strings = false;
		VolName_length_fixed = true;
		vol_number_32_bit = false;
		}
	else if(0 == strcmp(stringbuffer, "WSSI-63"))
		{
		bootable_flag_present = true;
		diskbits_one_byte = true;
		file_count_32_bit = true;
		file_size_64_bit = false;
		file_size_before_date = false;
		msdos_time_format = true;
		short_fn_ext_present = true;
		subdirs_one_disk_32_bit = false;
		unicode_strings = false;
		}
	else if(0 == strcmp(stringbuffer, "WSSI-64"))
		{
		bootable_flag_present = true;
		diskbits_one_byte = true;
		msdos_time_format = true;
		short_fn_ext_present = true;
		subdirs_one_disk_32_bit = false;
		unicode_strings = false;
		}
	else if(0 == strcmp(stringbuffer, "WSSINDEX"))
		{
		bootable_flag_present = true;
		diskbits_one_byte = true;
		disk_size_64_bit = false;
		file_count_32_bit = false;
		file_size_64_bit = false;
		file_size_before_date = false;
		msdos_time_format = true;
		old_format_no_categories = false;
		read_lfn = false;
		short_fn_ext_present = true;
		subdirs_one_disk_32_bit = false;
		total_file_count_32_bit = false;
		total_subdir_number_32_bit = false;
		unicode_strings = false;
		VolName_length_fixed = true;
		vol_number_32_bit = false;
		}
	else
		{
		QMessageBox::critical(this, tr("Error"),
			tr("%1 is not a WSSINDEX database!").arg(QDir::toNativeSeparators(database_name)));
		goto readclean;
		}
	if(!wgets(stringbuffer, 10, fp, false))
		goto readerr;
// stringbuffer has version
	old_format_no_categories = (*stringbuffer == '1');
	newdiskno = diskstart + getlong(fp, vol_number_32_bit ? READ_32_BITS : READ_16_BITS);
	newsubno = substart + getlong(fp, total_subdir_number_32_bit ? READ_32_BITS : READ_16_BITS);
	if(merging)
		newsubno--;
	newnofile = filestart + getlong(fp, total_file_count_32_bit ? READ_32_BITS : READ_16_BITS);
	if(WSSERROR_OR_EOF(fp))
		goto readerr;
	if(merging)
		{
		increase_max_filecount_to(qMax(static_cast<unsigned int>(Maxfiles_initial), newnofile+static_cast<unsigned int>(Maxfiles_increment)));
		increase_max_subcount_to(qMax(static_cast<unsigned int>(Maxsubs_initial), newsubno+static_cast<unsigned int>(Maxsubs_increment)));
		increase_max_diskcount_to(qMax(static_cast<unsigned int>(Maxdisks_initial), newdiskno+static_cast<unsigned int>(Maxdisks_increment)));
		}
	else
		{
		simpleconf(qMax(static_cast<unsigned int>(Maxfiles_initial), newnofile+static_cast<unsigned int>(Maxfiles_increment)),
				  qMax(static_cast<unsigned int>(Maxsubs_initial), newsubno+static_cast<unsigned int>(Maxsubs_increment)),
				  qMax(static_cast<unsigned int>(Maxdisks_initial), newdiskno+static_cast<unsigned int>(Maxdisks_increment)));
		}
	if((newdiskno>Maxdisks) || (newsubno>Maxsubs) || (newnofile>Maxfiles))
		{
		QMessageBox::critical(this, tr("Error"),
			tr("Database too large, requires %1 files, %2 subdirectories, %3 disks")
			.arg(newnofile).arg(newsubno).arg(newdiskno));
		goto readclean;
		}

	for(unsigned long i=diskstart; i<newdiskno; i++)
		{
		nvread = static_cast<long>(i);
		vrec_i = Vrec + i;
		if(unicode_strings)
			{
			if(!wgets(stringbuffer, STRINGBUFFER_SIZE, fp, false))
				{
				QMessageBox::critical(this, tr("Error"), tr("Error reading volume names"));
				goto readerr;
				}
			vrec_i->VolName = QString::fromUtf8(stringbuffer);
			}
		else if(VolName_length_fixed)
// no newline for fixed-length volume names, don't use wgets
			{
			if(!fgets(stringbuffer, 12, fp))
				goto readerr;
// This removes leading and trailing blanks; should perhaps only do trailing?
			vrec_i->VolName = QFA(stringbuffer).trimmed();
			}
		else
			{
			if(!wgets(stringbuffer, STRINGBUFFER_SIZE, fp, false))
				goto readerr;
			vrec_i->VolName = QFA(stringbuffer).trimmed();
			}

		if(disk_size_64_bit)
			{
			vrec_i->Vbytes = get64(fp);
			vrec_i->Vfree = get64(fp);
			}
		else
			{
			vrec_i->Vbytes = getlong(fp, READ_32_BITS);
			vrec_i->Vfree = getlong(fp, READ_32_BITS);
			}

		vrec_i->Vfiles = getlong(fp, file_count_32_bit ? READ_32_BITS : READ_16_BITS);
		vrec_i->Vsubdir = getlong(fp, subdirs_one_disk_32_bit ? READ_32_BITS : READ_16_BITS);

		if(msdos_time_format)
			vrec_i->Vdate = datextract(getshort(fp));
		else
			{
			int julday = static_cast<int>(getlong(fp, READ_32_BITS));
			vrec_i->Vdate = QDate::fromJulianDay(julday);
			}
		if(bootable_flag_present)
			getc(fp);		// skip obsolete Vboot
		}
// If merging, skip over subdirectory record for root
	if(merging)
		{
		getlong(fp, vol_number_32_bit ? READ_32_BITS : READ_16_BITS);
		if(!wgets(stringbuffer, 255, fp, false))
			goto readerr;
		}

// Classic databases have "\" as path separator; change to "/"
	for(unsigned long i=substart; i<newsubno; i++)
		{
		SubDisk[i] = diskstart + getlong(fp, vol_number_32_bit ? READ_32_BITS : READ_16_BITS);
		if(WSSERROR_OR_EOF(fp))
			goto readerr;
		if(!wgets(stringbuffer, 255, fp, false)) goto readerr;
		if(unicode_strings)
			SubDir[i] = QString::fromUtf8(stringbuffer);
		else
			SubDir[i] = QDir::fromNativeSeparators(stringbuffer);
		nsread = static_cast<long>(i);
		}

	last_good_vol = diskstart;
	flag_vol = newdiskno;
	prevVolNo = newdiskno;
	for(unsigned long i=filestart; i<newnofile; i++)
		{
		nfread = static_cast<long>(i);
		dr_temp = Dr + i;
		if(read_lfn)
			{
			if(diskbits_one_byte)
				dr_temp->Diskbits = static_cast<unsigned short>(getc(fp));
			else
				dr_temp->Diskbits = getshort(fp);
			}
		else
			dr_temp->Diskbits = 0;

		dr_temp->Bitflags = 0;
		if(hasComment(dr_temp))
			setHasComment(dr_temp);
		if(hasCategory(dr_temp))
			setHasCategory(dr_temp);
		if(hasExtraInfo(dr_temp))
			setHasExtra(dr_temp);

		struct short_fn
			{
			char	Dfn[9];			// file name, null terminated
			char	Dext[4];			// file extension, null terminated
			} old_short_name;

		if(short_fn_ext_present)
			if(fread(&old_short_name, 1, 13, fp) != 13) goto readerr;
		if(!file_size_before_date)
			{
			if(msdos_time_format)
				{
				dr_temp->Ddate = datextract(getshort(fp));
				dr_temp->Dtime = timextract(getshort(fp));
				}
			else
				{
				QMessageBox::critical(this, tr("Program error"),
					tr("Unexpected combination of date format and location of size value"));
				goto readerr;
				}
			}
		if(file_size_64_bit)
			dr_temp->Dsize = get64(fp);
		else
			dr_temp->Dsize = static_cast<unsigned long long>(getlong(fp, READ_32_BITS));
		if(file_size_before_date)
			{
			if(msdos_time_format)
				{
				dr_temp->Ddate = datextract(getshort(fp));
				dr_temp->Dtime = timextract(getshort(fp));
				}
			else
				{
				int julday = static_cast<int>(getlong(fp, READ_32_BITS));
				dr_temp->Ddate = QDate::fromJulianDay(julday);
				int tmsec = static_cast<int>(getlong(fp, READ_32_BITS));
				int ms = tmsec % 1000;
				tmsec /=1000;
				int sec = tmsec % 60;
				tmsec /= 60;
				int min = tmsec % 60;
				int hour = tmsec / 60;
				dr_temp->Dtime = QTime(hour, min, sec, ms);
				}
			}

		dr_temp->Volno = getlong(fp, vol_number_32_bit ? READ_32_BITS : READ_16_BITS);
		if(dr_temp->Volno & 0x80000000)
			{
			dr_temp->Volno &= 0x7fffffff;
			setInArc(dr_temp);
			}
		dr_temp->Subdirno = getlong(fp, total_subdir_number_32_bit ? READ_32_BITS : READ_16_BITS);

		if(merging)
			{
			dr_temp->Volno += diskstart;
			if(inArc(dr_temp))
				dr_temp->Subdirno += filestart;
			else if(dr_temp->Subdirno > 0)
				dr_temp->Subdirno += substart-1;
			}
		if(prevVolNo != dr_temp->Volno)
			{
//			qDebug() << "Set firstOnDisk = " << i << "for" << dr_temp->Volno;
			Vrec[dr_temp->Volno].firstOnDisk = i;
			prevVolNo = dr_temp->Volno;
			}
		if(!read_lfn && !inArc(dr_temp) && *old_short_name.Dfn == EOS)
			setIsDotID(dr_temp);

// check for subdirectory consistency
		if(consistent													// no error found yet
			&& !inArc(dr_temp)// not in archive file
			&& ((local_subdirno = dr_temp->Subdirno) > 0)	// not in root
			&& (local_subdirno >= newsubno
				|| SubDisk[local_subdirno] != dr_temp->Volno))	// inconsistent
			{
			local_volno = dr_temp->Volno;
			consistent = false;
			if(local_volno > newdiskno)
				local_volno = last_good_vol + 1;
			QMessageBox::critical(this, tr("Error"),
				tr("Inconsistent subdirectory info found starting at disk %1")
				.arg((Vrec + local_volno)->VolName));
			}

// check for valid disk number for archive members
		if(inArc(dr_temp))
			{
			if(dr_temp->Volno != last_good_vol)
				{
				if(vols_ok)
					{
					vols_ok = false;
					QMessageBox::critical(this, tr("Error"),
						tr("Invalid volume number found for archive file member(s)"));
					}
				dr_temp->Volno = last_good_vol;
// Check for valid subdirectory number which should actually be the file
//	number of the containing archive file.  If this points to garbage, an
//	archive member or a file on another disk, then move file to root, not
//	in archive and flag disk for reindexing.
				if(dr_temp->Subdirno > static_cast<unsigned>(nfread)
					|| ((dr_fix = Dr + dr_temp->Subdirno)->Volno)
					|| dr_fix->Volno != last_good_vol)
					{
					if(flag_vol != last_good_vol)
						{
						flag_vol = last_good_vol;
						vrec_i = Vrec + flag_vol;
						QMessageBox::information(this, tr("Warning"),
											tr("Please reindex %1")
											.arg(vrec_i->VolName));
						}
					dr_temp->Subdirno = 0;
					dr_temp->Volno = flag_vol;
					}
				}
			}
		if(static_cast<signed>(dr_temp->Volno) <= nvread)
			last_good_vol = dr_temp->Volno;

		dr_temp->Comment = QString();
		if(!read_lfn)
			{
			if((tempc = static_cast<unsigned char>(fgetc(fp))) == EOFCHAR)
				goto readerr;
			if('C' == tempc)
				setHasComment(dr_temp);
			}

		if(hasComment(dr_temp))
			{
			if(!wgets(stringbuffer, STRINGBUFFER_SIZE, fp, false)) goto readerr;
			if(EOS == *stringbuffer)
				setHasNoComment(dr_temp);
			else
				{
				setHasComment(dr_temp);
				if(unicode_strings)
					dr_temp->Comment = QString::fromUtf8(stringbuffer);
				else
					dr_temp->Comment = stringbuffer;
				}
			}
		dr_temp->Category = QString();
		dr_temp->Lfn = QString();
		if(!old_format_no_categories)
			{
			if(!read_lfn)
				{
				if((tempc = static_cast<unsigned char>(fgetc(fp))) == EOFCHAR)
					goto readerr;
				if('C' == tempc)
					setHasCategory(dr_temp);
				}

			if(hasCategory(dr_temp))
				{
				if(!wgets(stringbuffer, STRINGBUFFER_SIZE, fp, false))
					goto readerr;
// Bug corrected after version 5.13 would create 0-length categories if
//	Auto_category was enabled without specifying an Auto_category_string.
				if(EOS == *stringbuffer)
					setHasNoCategory(dr_temp);
				else
					{
					setHasCategory(dr_temp);
					if(unicode_strings)
						dr_temp->Category = QString::fromUtf8(stringbuffer);
					else
						dr_temp->Category = stringbuffer;
					}
				}

// ExtraInfo is new with Qt version, so no need for legacy tests
			if(hasExtraInfo(dr_temp))
				{
				if(!wgets(stringbuffer, STRINGBUFFER_SIZE, fp, false))
					goto readerr;
				dr_temp->ExtraInfo = stringbuffer;
				}

			if(hasLFN(dr_temp))
				{
				if(!wgets(stringbuffer, STRINGBUFFER_SIZE, fp, false))
					goto readerr;
				if(unicode_strings)
					dr_temp->Lfn = QString::fromUtf8(stringbuffer);
				else
					dr_temp->Lfn = stringbuffer;
//				qDebug() << "Lfn = " << dr_temp->Lfn;
				}
			else
				{
				char tname[13];
				strncpy(tname, old_short_name.Dfn, 9);
// Make sure null terminated even if old_short_name is malformed
				tname[8] = EOS;
				strcat(tname, ".");
				strncat(tname, old_short_name.Dext, 12);
				tname[12] = EOS;
				dr_temp->Lfn = tname;
				setHasLFN(dr_temp);
				}
			dr_temp->ExtOffset = static_cast<unsigned short>(lastdot(dr_temp->Lfn));
			}
// Volume and subdirectory records don't have valid time information
		if(isDotID(dr_temp))
			{
			dr_temp->Ddate = QDate();
			dr_temp->Dtime = QTime();
			}
		else
			NoPlainFiles++;
		}

	Diskno = newdiskno;
	Subno = newsubno;
	Nofile = newnofile;
	if(merging)
		purge_dups(true);
	statusChanged(merging);

	if(fclose(fp))
		{
		QMessageBox::critical(this, tr("Error"), tr("Error closing file"));
		}
	else
		free(iobuff);
//	if(warn_truncate)
//		{
//		QMessageBox::warning(this, tr("Warning"),
//						tr("Long comment(s) were found and truncated"));
//		}

	return true;

readerr:  // read errors come here
	QMessageBox::critical(this, tr("Error"),
		tr("Error reading database, nv, ns, nf = %1, %2, %3")
		.arg(nvread).arg(nsread).arg(nfread));
	Subno = qMax(1u, substart);
//	if(nsread > 0)
//		delete SubDir;	// ***** This is very wrong.  Only new subdirs should be deleted

	Diskno = diskstart;
	Nofile = filestart;

// Exit up after failing database read which did not damage in-core database.
// Loaded database may have been discarded, so model may still be invalid.
readclean:
	fclose(fp);
	free(iobuff);
	slot_updateDisplay(true);
	return false;

nocore:
	fclose(fp);
	QMessageBox::critical(this, tr("Error"),
		tr("Out of memory in routine getwss"));
	return false;
	}

