// This file is part of the disk cataloger Wssindex.  Wssindex is copyright
// 1985-2021 by Robert W. Babcock and WSS Division of DDC.  Effective 2010,
// Wssindex is distributed under the terms of the GNU General Public
// License.  Contact the author at wssddc@wssddc.com or see the project web
// page at <https://www.wssddc.com/wssindex/>.
//
// Wssindex uses the open source Qt toolkit from the Qt Project (qt-project.org)
// and The Qt Company (www.qt.io).
//
// Wssindex is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// Wssindex is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with Wssindex.  If not, see <https://www.gnu.org/licenses/>.

#include "OtherWssi/wssindex.h"
#include "Dialogs/mainwindow.h"
#include <stdio.h>
#include "OtherWssi/versioninfo.h"
#include "OtherWssi/globalsettings.h"
#include "Dialogs/fadeanimation.h"
#include "OtherWssi/wssiproto.h"

static bool put16(short number, FILE *fp);
static bool put32(long number, FILE *fp);
static bool put64(qint64 number, FILE *fp);

// putwss.c - write wssindex database on disk
// Returns true on success, false on failure
// (The various item writing routines use the opposite convention)

bool MainWindow::putwss(QString db_name)
	{
	FILE *fp;						// file pointer for output database
	struct filerec *dr_i;		// pointer to file record being saved
	struct diskrec *vrec_i;		// pointer to volume record being saved

// If db_name doesn't end with .dir, append same and warn user
	if(!db_name.endsWith(".dir", Qt::CaseInsensitive))
		{
// Avoid a clang memory leak warning, FadeAnimation deletes itself on close
#ifndef __clang_analyzer__
		new FadeAnimation(tr("Default extension set"), tr(".dir was appended to the database name."), Fade_time);
#endif
		db_name += ".dir";
		}
	std::wstring db_wide_name = db_name.toStdWString().data();

	QString bakname = db_name;
	bakname.replace(QRegExp("(\\.dir$)"), ".bak");
	std::wstring wide_backname = bakname.toStdWString().data();

	if((fp = _wfopen(db_wide_name.c_str(), L"r")))
		{
		fclose(fp);

// file already exists, if creating .BAK files, see if .BAK file also exists
		if(Makebak)
			{
			if(0 == QString::compare(db_name, bakname))
				{
				QMessageBox::critical(this, tr("Error"),
					tr("Something is wonky, backup file name is same as database name"));
				return false;
				}
			if((fp = _wfopen(bakname.toStdWString().data(), L"r")))
				{
// .BAK file exists, erase it
				fclose(fp);
				if(_wunlink(bakname.toStdWString().data()))
					{
					QMessageBox::critical(this, tr("Error"),
								tr("Unable to erase old .BAK file, new file not saved"));
					return false;
					}
				}
			if(_wrename(db_wide_name.c_str(), wide_backname.c_str()))
				{
				QMessageBox::critical(this, tr("Error"),
					tr("Unable to rename %1 as %2, new file not saved")
					.arg(QDir::toNativeSeparators(db_name), bakname));
				return false;
				}
			}
		else
			{
			int answer = QMessageBox::question(this, tr("Verify overwrite"),
				tr("%1 already exists and configuration is set to not make backup files.  OK to overwrite?")
									.arg(QDir::toNativeSeparators(db_name)), QMessageBox::Yes | QMessageBox::No);
			if(QMessageBox::Yes != answer)
				{
				QMessageBox::critical(this, tr("Action canceled"), tr("Database not saved"));
				return false;
				}
			}
		}

	if(!(fp = _wfopen(db_wide_name.c_str(), L"wb")))
		{
		QMessageBox::critical(this, tr("Error"),
			 tr("Unable to open file %1 for output").arg(QDir::toNativeSeparators(db_name)));
		return false;
		}

	if(wputs("WSSI-Qt", fp))
		goto writerr;

	if(wputs(WSSINDEX_VERSION, fp))
		goto writerr;
	if(put32(static_cast<long>(Diskno), fp))
		goto writerr;
	if(put32(static_cast<long>(Subno), fp))
		goto writerr;
	if(put32(static_cast<long>(Nofile), fp))
		goto writerr;
	for(unsigned long i=0; i<Diskno; i++)
		{
		vrec_i = Vrec + i;
		QByteArray vn_utf8 = vrec_i->VolName.toUtf8();
// now do want newline here, so do use wputs
		if(wputs(vn_utf8, fp))
			goto writerr;
		if(put64(static_cast<qint64>(vrec_i->Vbytes), fp))
			goto writerr;
		if(put64(static_cast<qint64>(vrec_i->Vfree), fp))
			goto writerr;
		if(put32(static_cast<long>(vrec_i->Vfiles), fp))
			goto writerr;
		if(put32(static_cast<long>(vrec_i->Vsubdir), fp))
			goto writerr;
		if(put32(static_cast<long>(vrec_i->Vdate.toJulianDay()), fp))
			goto writerr;
		}

// Subdirs are saved with "/" as path separator
	for(unsigned long i=0; i<Subno; i++)
		{
		if(put32(static_cast<long>(SubDisk[i]), fp))
			goto writerr;
		QByteArray subdir_utf16 = SubDir[i].toUtf8();
		if(wputs(subdir_utf16, fp))
			goto writerr;
		}

	for(unsigned long i=0; i<Nofile; i++)
		{
		dr_i = Dr + i;
		if(put16(static_cast<short>(dr_i->Diskbits), fp))
			goto writerr;
		if(put64(static_cast<qint64>(dr_i->Dsize), fp))
			goto writerr;
		if(isDotID(dr_i))
			put64(0, fp);
		else
			{
			if(put32(static_cast<long>(dr_i->Ddate.toJulianDay()), fp))
				goto writerr;
			if(put32(((dr_i->Dtime.hour()*60 + dr_i->Dtime.minute())*60
							+ dr_i->Dtime.second())*1000 + dr_i->Dtime.msec(), fp))
				goto writerr;
			}
		if(put32(static_cast<long>(dr_i->Volno), fp))
			goto writerr;
		if(put32(static_cast<long>(dr_i->Subdirno), fp))
			goto writerr;

		if(hasComment(dr_i))
			{
			QByteArray comment_utf16 = dr_i->Comment.toUtf8();
			if(wputs(comment_utf16, fp))
				goto writerr;
			}
// Following statement should not be inside the hasComment test because a
// deleted comment will have edited flag set but hasComment will be false.
		setCommentNotEdited(dr_i);
		if(hasCategory(dr_i))
			{
			QByteArray category_utf16 = dr_i->Category.toUtf8();
			if(wputs(category_utf16, fp))
				goto writerr;
			}
		setCategoryNotEdited(dr_i);
		if(hasExtraInfo(dr_i))
			{
			QByteArray extra_utf16 = dr_i->ExtraInfo.toUtf8();
			if(wputs(extra_utf16, fp))
				goto writerr;
			}
		setExtraNotEdited(dr_i);

		if(wputs(dr_i->Lfn.toUtf8(), fp))
			goto writerr;
		setVolNotEdited(dr_i);
		}
	if(fclose(fp))
		{
		QMessageBox::critical(this, tr("Error"), tr("Error closing file"));
		return false;
		}

// Avoid a clang memory leak warning, FadeAnimation deletes itself on close
#ifndef __clang_analyzer__
	new FadeAnimation(tr("Info"), tr("Save of %1 completed").arg(QDir::toNativeSeparators(db_name)), Fade_time);
#endif
	statusChanged(false);
	database_name = db_name;
	slot_updateDisplay(false);
	return true;

writerr:  // disk write errors come here
	QMessageBox::critical(this, tr("Error"), tr("Error writing file"));
	fclose(fp);
	slot_updateDisplay(false);
	return false;
	}

// put16 - write a short (16-bit) in the usual lsb first order.  Returns
//	true on error, false if ok.

static bool put16(short number, FILE *fp)
	{
	char *c;
	int i;

	c = reinterpret_cast<char *>(&number);
	for(i=0; i<2; i++)
		if(EOF == putc(*(c++), fp))
			return(true);

	return(false);
	}

// put32 - write a long (32-bit) in the usual lsb first order.  Returns
//	true on error, false if ok.

static bool put32(long number, FILE *fp)
	{
	char *c;
	int i;

	c = reinterpret_cast<char *>(&number);
	for(i=0; i<4; i++)
		if(EOF == putc(*(c++), fp))
			return(true);

	return(false);
	}

// put64 - write a 64-bit integer in the usual lsb first order.  Returns
//	true on error, false if ok.

static bool put64(qint64 number, FILE *fp)
	{
	char *c;
	int i;

	c = reinterpret_cast<char *>(&number);
	for(i=0; i<8; i++)
		if(EOF == putc(*(c++), fp))
			return(true);

	return(false);
	}
